import unittest
import numpy as np


class TestCoercion(unittest.TestCase):
    def test_as_numeric_array_three_by_one(self):
        from .coercion import as_numeric_array

        good = [np.array([1, 2, 3])]

        good_iff_allow_none = [None]

        good_iff_allow_none_and_empty_as_none = [tuple(), []]

        bad = [
            (1,),
            (1, 2),
            (1, 2, 3, 4),
            [1],
            [1, 2],
            [1, 2, 3, 4],
            np.array([1, 2]),
            np.array([1, 2, 3, 4]),
            np.array([1, 2, 3]).reshape(3, 1),
            ("foo", "bar", "baz"),
            ["foo", "bar", "baz"],
        ]

        for v in good:
            as_numeric_array(v, shape=(3,), allow_none=True)
            as_numeric_array(v, shape=(3,), allow_none=False)

        for v in good_iff_allow_none:
            as_numeric_array(v, shape=(3,), allow_none=True)
            with self.assertRaises(ValueError):
                as_numeric_array(v, shape=(3,), allow_none=False)

        for v in good_iff_allow_none_and_empty_as_none:
            as_numeric_array(v, shape=(3,), allow_none=True, empty_as_none=True)
            with self.assertRaises(ValueError):
                as_numeric_array(v, shape=(3,), allow_none=False, empty_as_none=True)

        for v in bad:
            with self.assertRaises(ValueError):
                as_numeric_array(v, shape=(3,), allow_none=True)
            with self.assertRaises(ValueError):
                as_numeric_array(v, shape=(3,), allow_none=False)



import numpy as np


def estimate_normal(planar_points):
    from sklearn.decomposition import PCA

    pca = PCA(n_components=3)
    normal = pca.fit(planar_points).components_[-1]
    normal /= np.linalg.norm(normal)
    return normal


def rotate_to_xz_plane(points, normal=None):
    """
    Rotates points to the x-z plane. If the initial center
    of mass is not to within 1e-5 of the origin, we
    translate it to the origin.

    Returns (r, R, p0):

        - r is the rotated image of points,
        - R is the rotation matrix
        - p0 is the translation factor (can be None)

    """
    import cv2
    from .translation import translation

    if points is None or not len(points):  # pylint: disable=len-as-condition
        raise ValueError("Some points are required")

    center = np.mean(points, axis=0)

    if np.linalg.norm(center) > 1e-5:
        translated, p0 = translation(points)
    else:
        translated, p0 = points, None

    if not normal:
        normal = estimate_normal(points)

    e_2 = np.array([0.0, 1.0, 0.0])
    theta = np.arccos(np.dot(e_2, normal))

    if min(abs(theta - np.pi), abs(theta)) < 1e-5:
        # cross product will degenerate
        # to zero vector in this case
        r_axis = np.array([1.0, 0.0, 0.0])
    else:
        r_axis = np.cross(normal, e_2)

    r_axis /= np.linalg.norm(r_axis)

    R = cv2.Rodrigues(theta * r_axis)[0]
    rotated = np.dot(translated, R.T)

    return (rotated, R, p0)


def rotation_from_up_and_look(up, look):
    """
    Rotation matrix to rotate a mesh into a canonical reference frame. The
    result is a rotation matrix that will make up along +y and look along +z
    (i.e. facing towards a default opengl camera).

    Note that if you're reorienting a mesh, you can use its `reorient` method
    to accomplish this.

    up: The foot-to-head direction.
    look: The direction the eyes are facing, or the heel-to-toe direction.

    """
    up, look = np.array(up, dtype=np.float64), np.array(look, dtype=np.float64)
    if np.linalg.norm(up) == 0:
        raise ValueError("Singular up")
    if np.linalg.norm(look) == 0:
        raise ValueError("Singular look")
    y = up / np.linalg.norm(up)
    z = look - np.dot(look, y) * y
    if np.linalg.norm(z) == 0:
        raise ValueError("up and look are colinear")
    z = z / np.linalg.norm(z)
    x = np.cross(y, z)
    return np.array([x, y, z])


def euler(xyz, order="xyz", units="deg"):
    if not hasattr(xyz, "__iter__"):
        xyz = [xyz]
    if units == "deg":
        xyz = np.radians(xyz)
    r = np.eye(3)
    for theta, axis in zip(xyz, order):
        c = np.cos(theta)
        s = np.sin(theta)
        if axis == "x":
            r = np.dot(np.array([[1, 0, 0], [0, c, -s], [0, s, c]]), r)
        if axis == "y":
            r = np.dot(np.array([[c, 0, s], [0, 1, 0], [-s, 0, c]]), r)
        if axis == "z":
            r = np.dot(np.array([[c, -s, 0], [s, c, 0], [0, 0, 1]]), r)
    return r



import numpy as np
import vg


def convert_33_to_44(matrix):
    """
    Transform from:
        array([[1., 2., 3.],
               [2., 3., 4.],
               [5., 6., 7.]])
    to:
        array([[1., 2., 3., 0.],
               [2., 3., 4., 0.],
               [5., 6., 7., 0.],
               [0., 0., 0., 1.]])

    """
    if matrix.shape != (3, 3):
        raise ValueError("Expected 3x3 matrix, got: %s" % matrix.shape)
    result = np.pad(matrix, ((0, 1), (0, 1)), mode="constant")
    result[3][3] = 1
    return result


def convert_44_to_33(matrix):
    """
    Transform from:
        array([[1., 2., 3., 0.],
               [2., 3., 4., 0.],
               [5., 6., 7., 0.],
               [0., 0., 0., 1.]])
    to:
        array([[1., 2., 3.],
               [2., 3., 4.],
               [5., 6., 7.]])

    """
    if matrix.shape != (4, 4):
        raise ValueError("Expected 4x4 matrix, got: %s" % matrix.shape)
    return matrix[:3, :3]


class CompositeTransform(object):
    """
    Composite transform using homogeneous coordinates.


    Example usage
    -------------

        transform = CompositeTransform()
        transform.scale(10)
        transform.reorient(up=[0, 1, 0], look=[-1, 0, 0])
        transform.translate([0, -2.5, 0])
        transformed_scan = Mesh(v=transform(scan.v), f=scan.f)

        ...

        untransformed_alignment = Mesh(
            v=transform(alignment.v, reverse=True),
            f=alignment.f
        )

    Backround
    ---------

    - Computer Graphics: Principles and Practice, Hughes, van Dam, McGuire,
      Sklar, Foley

    - http://gamedev.stackexchange.com/questions/72044/why-do-we-use-4x4-matrices-to-transform-things-in-3d

    """

    def __init__(self):
        # List of tuples, containing forward and reverse matrices.
        self.transforms = []

    def __call__(self, points, from_range=None, reverse=False):
        """
        points: Points to transform, as a 3xn array.
        range: The indices of the subset of the transformations to apply.
          e.g. (0, 2), (2, 4). The default is to apply them all.
        reverse: When `True` applies the selected transformations in reverse.
          This has no effect on how range is interpreted, only whether the
          selected transformations apply in the forward or reverse mode.

        """
        matrix = self.matrix_for(from_range=from_range, reverse=reverse)

        return vg.unpad(np.dot(matrix, vg.pad_with_ones(points).T).T)

    def matrix_for(self, from_range=None, reverse=False):
        """
        Return a 4x4 transformation matrix representation.

        range: The min and max indices of the subset of the transformations to
          apply. e.g. (0, 2), (2, 4). Inclusive of the min value, exclusive of
          the max value. The default is to apply them all.
        reverse: When `True` returns a matrix for the inverse transform.
          This has no effect on how range is interpreted, only whether the
          forward or reverse matrices are used.

        """
        from functools import reduce

        if from_range is not None:
            start, stop = (
                from_range
            )  # from_range is defined as None, a non-sequence, but when it's not None, it's always a sequence. pylint: disable=unpacking-non-sequence
            selected_transforms = self.transforms[start:stop]
        else:
            selected_transforms = self.transforms

        # The transpose of a product of matrices equals the products of each
        # transpose in reverse order.
        matrices = [
            reverse_matrix if reverse else forward_matrix.T
            for forward_matrix, reverse_matrix in selected_transforms
        ]

        if not len(matrices):  # pylint: disable=len-as-condition
            return np.eye(4)

        matrix = reduce(np.dot, matrices)
        return matrix if reverse else matrix.T

    def append_transform4(self, forward, reverse=None):
        """
        Append an arbitrary transformation, defined by 4x4 forward and reverse
        matrices.

        The new transformation is added to the end. Return its index.

        """
        if reverse is None:
            reverse = np.linalg.inv(forward)

        new_index = len(self.transforms)

        self.transforms.append((forward, reverse))

        return new_index

    def append_transform3(self, forward, reverse=None):
        """
        Append an arbitrary transformation, defined by 3x3 forward and reverse
        matrices.

        The new transformation is added to the end. Return its index.

        """
        args = (forward,) if reverse is None else (forward, reverse)
        return self.append_transform4(*map(convert_33_to_44, args))

    def scale(self, factor):
        """
        Scale by the given factor.

        factor: A float or int.

        Forward:
        [[  s_0, 0,   0,   0 ],
         [  0,   s_1, 0,   0 ],
         [  0,   0,   s_2, 0 ],
         [  0,   0,   0,   1 ]]

        Reverse:
        [[  1/s_0, 0,     0,     0 ],
         [  0,     1/s_1, 0,     0 ],
         [  0,     0,     1/s_2, 0 ],
         [  0,     0,     0,     1 ]]

        """
        if factor <= 0:
            raise ValueError("Scale factor should be greater than zero")

        forward3 = np.diag(np.repeat(factor, 3))
        reverse3 = np.diag(np.repeat(1.0 / factor, 3))

        return self.append_transform3(forward3, reverse3)

    def convert_units(self, from_units, to_units):
        """
        Convert the mesh from one set of units to another.

        These calls are equivalent:

        - composite.convert_units(from_units='cm', to_units='m')
        - composite.scale(.01)

        """
        from .._temporary import units

        factor = units.factor(
            from_units=from_units, to_units=to_units, units_class="length"
        )
        self.scale(factor)

    def translate(self, vector):
        """
        Translate by the vector provided.

        vector: A 3x1 vector.

        Forward:
        [[  1,  0,  0,  v_0 ],
         [  0,  1,  0,  v_1 ],
         [  0,  0,  1,  v_2 ],
         [  0,  0,  0,  1   ]]

        Reverse:
        [[  1,  0,  0,  -v_0 ],
         [  0,  1,  0,  -v_1 ],
         [  0,  0,  1,  -v_2 ],
         [  0,  0,  0,  1    ]]

        """
        vector = np.asarray(vector)

        forward = np.eye(4)
        forward[:, -1][:-1] = vector

        reverse = np.eye(4)
        reverse[:, -1][:-1] = -vector

        return self.append_transform4(forward, reverse)

    def reorient(self, up, look):
        """
        Reorient using up and look.

        """
        from .rotation import rotation_from_up_and_look

        forward3 = rotation_from_up_and_look(up, look)
        # The inverse of a rotation matrix is its transpose.
        return self.append_transform3(forward3, forward3.T)

    def rotate(self, rot):
        """
        Rotate by either an explicit matrix or a rodrigues vector
        """
        from .rodrigues import as_rotation_matrix

        rot = np.asarray(rot)
        rot = as_rotation_matrix(rot)
        # The inverse of a rotation matrix is its transpose.
        return self.append_transform3(rot, rot.T)



import numpy as np
import vg
from .._temporary.decorators import setter_property


class Polyline(object):
    """
    Represent the geometry of a polygonal chain in 3-space. The
    chain may be open or closed, and there are no constraints on the
    geometry. For example, the chain may be simple or
    self-intersecting, and the points need not be unique.

    Mutable by setting polyline.v or polyline.closed or calling
    a method like polyline.partition_by_length().

    Note this class is distinct from lace.lines.Lines, which
    allows arbitrary edges and enables visualization. To convert to
    a Lines object, use the as_lines() method.

    """

    def __init__(self, v, closed=False):
        """
        v: An array-like thing containing points in 3-space.
        closed: True indicates a closed chain, which has an extra
          segment connecting the last point back to the first
          point.

        """
        # Avoid invoking _update_edges before setting closed and v.
        self.__dict__["closed"] = closed
        self.v = v

    def __repr__(self):
        if self.v is not None and len(self.v) != 0:
            if self.closed:
                return "<closed Polyline with {} verts>".format(len(self))
            else:
                return "<open Polyline with {} verts>".format(len(self))
        else:
            return "<Polyline with no verts>"

    def __len__(self):
        return len(self.v)

    def copy(self):
        """
        Return a copy of this polyline.

        """
        v = None if self.v is None else np.copy(self.v)
        return self.__class__(v, closed=self.closed)

    def as_lines(self, vc=None):
        """
        Return a Lines instance with our vertices and edges.

        """
        from lace.lines import Lines

        return Lines(v=self.v, e=self.e, vc=vc)

    def to_dict(self, decimals=3):
        return {
            "vertices": [np.around(v, decimals=decimals).tolist() for v in self.v],
            "edges": self.e,
        }

    def _update_edges(self):
        if self.v is None:
            self.__dict__["e"] = None
            return

        num_vertices = self.v.shape[0]
        num_edges = num_vertices if self.closed else num_vertices - 1

        edges = np.vstack([np.arange(num_edges), np.arange(num_edges) + 1]).T

        if self.closed:
            edges[-1][1] = 0

        edges.flags.writeable = False

        self.__dict__["e"] = edges

    @setter_property
    def v(
        self, val
    ):  # setter_property incorrectly triggers method-hidden. pylint: disable=method-hidden
        """
        Update the vertices to a new array-like thing containing points
        in 3D space. Set to None for an empty polyline.

        """
        from .._temporary.coercion import as_numeric_array

        self.__dict__["v"] = as_numeric_array(
            val, dtype=np.float64, shape=(-1, 3), allow_none=True
        )
        self._update_edges()

    @setter_property
    def closed(self, val):
        """
        Update whether the polyline is closed or open.

        """
        self.__dict__["closed"] = val
        self._update_edges()

    @property
    def e(self):
        """
        Return a np.array of edges. Derived automatically from self.v
        and self.closed whenever those values are set.

        """
        return self.__dict__["e"]

    @property
    def segments(self):
        """
        Coordinate pairs for each segment.
        """
        return self.v[self.e]

    @property
    def segment_lengths(self):
        """
        The length of each of the segments.

        """
        if self.e is None:
            return np.empty((0,))

        return ((self.v[self.e[:, 1]] - self.v[self.e[:, 0]]) ** 2.0).sum(axis=1) ** 0.5

    @property
    def total_length(self):
        """
        The total length of all the segments.

        """
        return np.sum(self.segment_lengths)

    @property
    def segment_vectors(self):
        """
        Vectors spanning each segment.
        """
        segments = self.segments
        return segments[:, 1] - segments[:, 0]

    def flip(self):
        """
        Flip the polyline from end to end.
        """
        self.v = np.flipud(self.v)

    def reindexed(self, index, ret_edge_mapping=False):
        """
        Return a new Polyline which reindexes the callee polyline, which much
        be closed, so the vertex with the given index becomes vertex 0.

        ret_edge_mapping: if True, return an array that maps from old edge
            indices to new.
        """
        if not self.closed:
            raise ValueError("Can't reindex an open polyline")

        result = Polyline(
            v=np.append(self.v[index:], self.v[0:index], axis=0), closed=True
        )

        if ret_edge_mapping:
            edge_mapping = np.append(
                np.arange(index, len(self.v)), np.arange(0, index), axis=0
            )
            return result, edge_mapping
        else:
            return result

    def partition_by_length(self, max_length, ret_indices=False):
        """
        Subdivide each line segment longer than max_length with
        equal-length segments, such that none of the new segments
        are longer than max_length.

        ret_indices: If True, return the indices of the original vertices.
          Otherwise return self for chaining.

        """
        from ..segment.segment import partition_segment

        lengths = self.segment_lengths
        num_segments_needed = np.ceil(lengths / max_length)

        indices_of_orig_vertices = []
        new_v = np.empty((0, 3))

        for i, num_segments in enumerate(num_segments_needed):
            start_point, end_point = self.v[self.e[i]]

            indices_of_orig_vertices.append(len(new_v))

            # In the simple case, one segment, or degenerate case, with
            # a repeated vertex, we do not need to subdivide.
            if num_segments <= 1:
                new_v = np.vstack((new_v, start_point.reshape(-1, 3)))
            else:
                new_v = np.vstack(
                    (
                        new_v,
                        partition_segment(
                            start_point, end_point, np.int(num_segments), endpoint=False
                        ),
                    )
                )

        if not self.closed:
            indices_of_orig_vertices.append(len(new_v))
            new_v = np.vstack((new_v, self.v[-1].reshape(-1, 3)))

        self.v = new_v

        return np.array(indices_of_orig_vertices) if ret_indices else self

    def apex(self, axis):
        """
        Find the most extreme point in the direction of the axis provided.

        axis: A vector, which is an 3x1 np.array.

        """
        return vg.apex(self.v, axis)

    def intersect_plane(self, plane, ret_edge_indices=False):
        """
        Returns the points of intersection between the plane and any of the
        edges of `polyline`, which should be an instance of Polyline.
        """
        # Identify edges with endpoints that are not on the same side of the plane
        sgn_dists = plane.signed_distance(self.v)
        which_es = np.abs(np.sign(sgn_dists)[self.e].sum(axis=1)) != 2
        # For the intersecting edges, compute the distance of the endpoints to the plane
        endpoint_distances = np.abs(sgn_dists[self.e[which_es]])
        # Normalize the rows of endpoint_distances
        t = endpoint_distances / endpoint_distances.sum(axis=1)[:, np.newaxis]
        # Take a weighted average of the endpoints to obtain the points of intersection
        intersection_points = (
            (1.0 - t[:, :, np.newaxis]) * self.segments[which_es]
        ).sum(axis=1)
        if ret_edge_indices:
            return intersection_points, which_es.nonzero()[0]
        else:
            return intersection_points

    def _cut_by_plane_open(self, plane, intersection_point, intersecting_edge_index):
        intersecting_edge = self.e[intersecting_edge_index]
        intersecting_segment_vector = self.segment_vectors[intersecting_edge_index]
        # Do we want the edges before or after the intersecting edge? Determine that
        # by checking whether the intersecting edge crosses from back to front or
        # front to back. If back to front, keep the edges after it. If front to back,
        # keep the edges before it. In either case, add a new edge for the portion of
        # the intersecting edge that is in front.
        #
        # In case a vert of the intersecting edge lies on the plane, use a
        # vector to identify which direction it's facing.
        if vg.sproj(intersecting_segment_vector, onto=plane.normal) > 0:
            new_v = np.vstack([intersection_point, self.v[intersecting_edge[1] :]])
        else:
            new_v = np.vstack([self.v[: intersecting_edge[0] + 1], intersection_point])
        return Polyline(v=new_v, closed=False)

    def _cut_by_plane_closed(self, plane):
        """
        Precondition: there are exactly two points of intersection.
        """
        # Reindex the polyline so it starts with a contiguous subchain that
        # lies in front of or on the plane.
        v_sign = plane.sign(self.v)
        first_point_is_in_back = v_sign[0] == -1
        if first_point_is_in_back:
            # e.g. v_sign = np.array([-1, 1, 1, 1, -1, -1, -1])
            points_on_or_in_front, = np.where(v_sign >= 0)
            working = self.reindexed(index=points_on_or_in_front[0])
        else:
            # e.g. v_sign = np.array([1, -1, -1, -1, 1, 1, 1])
            points_in_back, = np.where(v_sign < 0)
            working = self.reindexed(index=points_in_back[-1] + 1)

        # The part of the polyline we want to keep starts somewhere between
        # the first vertex, which lies on the front, and the previous vertex,
        # which lies on the back. Compute the starting point, which is the
        # point of intersection. The exception is if `working.v[0]` happens to
        # lie on the plane, in which case we don't need to synthesize a
        # starting point.
        v_sign = plane.sign(working.v)
        if v_sign[0] == 0:
            synthesized_start_point = None
        else:
            synthesized_start_point = plane.line_segment_xsection(*working.segments[-1])

        # Repeat for the endpoint of the part we want to keep.
        points_on_or_in_front, = np.where(v_sign >= 0)
        end_index = points_on_or_in_front[-1]
        if v_sign[end_index] == 0:
            synthesized_end_point = None
        else:
            synthesized_end_point = plane.line_segment_xsection(
                *working.segments[end_index]
            )

        points_to_keep = working.v[0 : end_index + 1]

        def to_array(maybe_point):
            if maybe_point is None:
                return np.zeros((0, 3))
            else:
                return np.array([maybe_point])

        return Polyline(
            v=np.concatenate(
                [
                    to_array(synthesized_start_point),
                    points_to_keep,
                    to_array(synthesized_end_point),
                ]
            ),
            closed=False,
        )

    def cut_by_plane(self, plane):
        """
        Return a new Polyline which keeps only the part that is in front of the given
        plane.

        For open polylines, the plane must intersect the polyline exactly once.

        For closed polylines, the plane must intersect the polylint exactly
        twice, leaving a single contiguous segment in front.
        """
        intersection_points, edge_indices = self.intersect_plane(
            plane, ret_edge_indices=True
        )
        num_edge_indices = len(edge_indices)
        if num_edge_indices == 0:
            raise ValueError("Plane does not intersect polyline")
        if self.closed:
            if num_edge_indices != 2:
                raise ValueError(
                    "Plane intersects polyline at {} points; expected 2".format(
                        num_edge_indices
                    )
                )
            return self._cut_by_plane_closed(plane)
        else:
            if num_edge_indices > 1:
                raise ValueError(
                    "Plane intersects polyline at {} points; expected 1".format(
                        num_edge_indices
                    )
                )
            intersection_point = intersection_points[0]
            intersecting_edge_index = edge_indices[0]
            return self._cut_by_plane_open(
                plane, intersection_point, intersecting_edge_index
            )



import numpy as np


def rodrigues(r, calculate_jacobian=False):
    r = np.array(r, dtype=np.double)
    eps = np.finfo(np.double).eps

    if (
        np.all(r.shape == (3, 1))
        or np.all(r.shape == (1, 3))
        or np.all(r.shape == (3,))
    ):
        r = r.flatten()
        theta = np.linalg.norm(r)
        if theta < eps:
            r_out = np.eye(3)
            if calculate_jacobian:
                jac = np.zeros((3, 9))
                jac[0, 5] = jac[1, 6] = jac[2, 1] = -1
                jac[0, 7] = jac[1, 2] = jac[2, 3] = 1

        else:
            c = np.cos(theta)
            s = np.sin(theta)
            c1 = 1.0 - c
            itheta = 1.0 if theta == 0.0 else 1.0 / theta
            r *= itheta
            I = np.eye(3)
            rrt = np.array([r * r[0], r * r[1], r * r[2]])
            _r_x_ = np.array([[0, -r[2], r[1]], [r[2], 0, -r[0]], [-r[1], r[0], 0]])
            r_out = c * I + c1 * rrt + s * _r_x_
            if calculate_jacobian:
                drrt = np.array(
                    [
                        [r[0] + r[0], r[1], r[2], r[1], 0, 0, r[2], 0, 0],
                        [0, r[0], 0, r[0], r[1] + r[1], r[2], 0, r[2], 0],
                        [0, 0, r[0], 0, 0, r[1], r[0], r[1], r[2] + r[2]],
                    ]
                )
                d_r_x_ = np.array(
                    [
                        [0, 0, 0, 0, 0, -1, 0, 1, 0],
                        [0, 0, 1, 0, 0, 0, -1, 0, 0],
                        [0, -1, 0, 1, 0, 0, 0, 0, 0],
                    ]
                )
                I = np.array([I.flatten(), I.flatten(), I.flatten()])
                ri = np.array([[r[0]], [r[1]], [r[2]]])
                a0 = -s * ri
                a1 = (s - 2 * c1 * itheta) * ri
                a2 = np.ones((3, 1)) * c1 * itheta
                a3 = (c - s * itheta) * ri
                a4 = np.ones((3, 1)) * s * itheta
                jac = (
                    a0 * I
                    + a1 * rrt.flatten()
                    + a2 * drrt
                    + a3 * _r_x_.flatten()
                    + a4 * d_r_x_
                )
    elif np.all(r.shape == (3, 3)):
        u, _, v = np.linalg.svd(r)
        r = np.dot(u, v)
        rx = r[2, 1] - r[1, 2]
        ry = r[0, 2] - r[2, 0]
        rz = r[1, 0] - r[0, 1]
        s = np.linalg.norm(np.array([rx, ry, rz])) * np.sqrt(0.25)
        c = np.clip((np.sum(np.diag(r)) - 1) * 0.5, -1, 1)
        theta = np.arccos(c)
        if s < 1e-5:
            if c > 0:
                r_out = np.zeros((3, 1))
            else:
                rx, ry, rz = np.clip(np.sqrt((np.diag(r) + 1) * 0.5), 0, np.inf)
                if r[0, 1] < 0:
                    ry = -ry
                if r[0, 2] < 0:
                    rz = -rz
                if (
                    np.abs(rx) < np.abs(ry)
                    and np.abs(rx) < np.abs(rz)
                    and ((r[1, 2] > 0) != (ry * rz > 0))
                ):
                    rz = -rz

                r_out = np.array([[rx, ry, rz]]).T
                theta /= np.linalg.norm(r_out)
                r_out *= theta
            if calculate_jacobian:
                jac = np.zeros((9, 3))
                if c > 0:
                    jac[1, 2] = jac[5, 0] = jac[6, 1] = -0.5
                    jac[2, 1] = jac[3, 2] = jac[7, 0] = 0.5
        else:
            vth = 1.0 / (2.0 * s)
            if calculate_jacobian:
                dtheta_dtr = -1.0 / s
                dvth_dtheta = -vth * c / s
                d1 = 0.5 * dvth_dtheta * dtheta_dtr
                d2 = 0.5 * dtheta_dtr
                dvardR = np.array(
                    [
                        [0, 0, 0, 0, 0, 1, 0, -1, 0],
                        [0, 0, -1, 0, 0, 0, 1, 0, 0],
                        [0, 1, 0, -1, 0, 0, 0, 0, 0],
                        [d1, 0, 0, 0, d1, 0, 0, 0, d1],
                        [d2, 0, 0, 0, d2, 0, 0, 0, d2],
                    ]
                )
                dvar2dvar = np.array(
                    [
                        [vth, 0, 0, rx, 0],
                        [0, vth, 0, ry, 0],
                        [0, 0, vth, rz, 0],
                        [0, 0, 0, 0, 1],
                    ]
                )
                domegadvar2 = np.array(
                    [
                        [theta, 0, 0, rx * vth],
                        [0, theta, 0, ry * vth],
                        [0, 0, theta, rz * vth],
                    ]
                )
                jac = np.dot(np.dot(domegadvar2, dvar2dvar), dvardR)
                for ii in range(3):
                    jac[ii] = jac[ii].reshape((3, 3)).T.flatten()
                jac = jac.T
            vth *= theta
            r_out = np.array([[rx, ry, rz]]).T * vth
    else:
        raise Exception("rodrigues: input matrix must be 1x3, 3x1 or 3x3.")
    if calculate_jacobian:
        return r_out, jac
    else:
        return r_out


def as_rotation_matrix(r):
    if (
        np.all(r.shape == (3, 1))
        or np.all(r.shape == (1, 3))
        or np.all(r.shape == (3,))
    ):
        return rodrigues(r)
    elif np.all(r.shape == (3, 3)):
        return rodrigues(rodrigues(r))
    else:
        raise Exception("rodrigues: input matrix must be 1x3, 3x1 or 3x3.")



import numpy as np


def _maybe_flatten(vertices, faces, ret_unique_vertices_and_faces):
    if ret_unique_vertices_and_faces:
        return vertices, faces
    else:
        return vertices[faces]


def create_rectangular_prism(origin, size, ret_unique_vertices_and_faces=False):
    """
    Return vertices (or unique verties and faces) of an axis-aligned
    rectangular prism. One vertex is `origin`; the diametrically opposite
    vertex is `origin + size`.

    size: 3x1 array.

    """
    from lace.topology import quads_to_tris

    lower_base_plane = np.array(
        [
            # Lower base plane
            origin,
            origin + np.array([size[0], 0, 0]),
            origin + np.array([size[0], 0, size[2]]),
            origin + np.array([0, 0, size[2]]),
        ]
    )
    upper_base_plane = lower_base_plane + np.array([0, size[1], 0])

    vertices = np.vstack([lower_base_plane, upper_base_plane])

    faces = quads_to_tris(
        np.array(
            [
                [0, 1, 2, 3],  # lower base (-y)
                [7, 6, 5, 4],  # upper base (+y)
                [4, 5, 1, 0],  # +z face
                [5, 6, 2, 1],  # +x face
                [6, 7, 3, 2],  # -z face
                [3, 7, 4, 0],  # -x face
            ]
        )
    )

    return _maybe_flatten(vertices, faces, ret_unique_vertices_and_faces)


def create_cube(origin, size, ret_unique_vertices_and_faces=False):
    """
    Return vertices (or unique verties and faces) with an axis-aligned cube.
    One vertex is `origin`; the diametrically opposite vertex is `size` units
    along +x, +y, and +z.

    size: int or float.

    """
    return create_rectangular_prism(
        origin,
        np.repeat(size, 3),
        ret_unique_vertices_and_faces=ret_unique_vertices_and_faces,
    )


def create_triangular_prism(p1, p2, p3, height, ret_unique_vertices_and_faces=False):
    """
    Return vertices (or unique verties and faces) of a triangular prism whose
    base is the triangle p1, p2, p3. If the vertices are oriented in a
    counterclockwise direction, the prism extends from behind them.

    Imported from lace.
    """
    from ..plane.plane import Plane

    base_plane = Plane.from_points(p1, p2, p3)
    lower_base_to_upper_base = (
        height * -base_plane.normal
    )  # pylint: disable=invalid-unary-operand-type
    vertices = np.vstack(([p1, p2, p3], [p1, p2, p3] + lower_base_to_upper_base))

    faces = np.array(
        [
            [0, 1, 2],  # base
            [0, 3, 4],
            [0, 4, 1],  # side 0, 3, 4, 1
            [1, 4, 5],
            [1, 5, 2],  # side 1, 4, 5, 2
            [2, 5, 3],
            [2, 3, 0],  # side 2, 5, 3, 0
            [5, 4, 3],  # base
        ]
    )

    return _maybe_flatten(vertices, faces, ret_unique_vertices_and_faces)


def create_horizontal_plane(ret_unique_vertices_and_faces=False):
    """
    Creates a horizontal plane.
    """
    vertices = np.array(
        [[1.0, 0.0, 0.0], [-1.0, 0.0, 0.0], [0.0, 0.0, 1.0], [0.0, 0.0, -1.0]]
    )
    faces = [[0, 1, 2], [3, 1, 0]]
    return _maybe_flatten(vertices, faces, ret_unique_vertices_and_faces)



class CoordinateManager(object):
    """
    Here's the idea:

        coordinate_manager = CoordinateManager()
        coordinate_manager.tag_as('source')
        coordinate_manager.translate(-cube.floor_point)
        coordinate_manager.scale(2)
        coordinate_manager.tag_as('floored_and_scaled')
        coordinate_manager.translate(np.array([0., -4., 0.]))
        coordinate_manager.tag_as('centered_at_origin')

        coordinate_manager.source = cube
        centered_mesh = coordinate_manager.centered_at_origin

    """

    def __init__(self):
        from .composite import CompositeTransform

        self.__dict__.update(
            {
                # A map from tag names to indices into the transform stack.
                "tags_to_indices": {},
                # Our currently set points, and the tag at which they belong.
                "points_tag": None,
                "points": None,
                # Our worthy collaborator.
                "_transform": CompositeTransform(),
            }
        )

    def append_transform4(self, *args, **kwargs):
        self._transform.append_transform4(*args, **kwargs)

    def append_transform3(self, *args, **kwargs):
        self._transform.append_transform3(*args, **kwargs)

    def scale(self, *args, **kwargs):
        self._transform.scale(*args, **kwargs)

    def convert_units(self, *args, **kwargs):
        self._transform.convert_units(*args, **kwargs)

    def translate(self, *args, **kwargs):
        self._transform.translate(*args, **kwargs)

    def reorient(self, *args, **kwargs):
        self._transform.reorient(*args, **kwargs)

    def rotate(self, *args, **kwargs):
        self._transform.rotate(*args, **kwargs)

    def tag_as(self, name):
        """
        Give a name to the current state.

        """
        # The indices of CompositeTransform are 0-based, which means the first
        # transform is transform 0.
        #
        # In CoordinateManager, we refer to the initial state -- i.e. no
        # transforms -- with 0. The state after the first transform is 1.
        # After two transforms, 2. Put another way: we refer to a given state
        # with a value that is *one more than* the index of the state's last
        # transform. This is a little strange, but we need the extra zero
        # state. And it ends up playing nicely with array slicing.
        self.tags_to_indices[name] = len(self._transform.transforms)

    def do_transform(self, points_or_mesh, from_tag, to_tag):
        from copy import copy

        if hasattr(points_or_mesh, "v"):
            points = points_or_mesh.v
            # Can't run the transform if there are no vertices.
            if points is None:
                return points_or_mesh
        else:
            points = points_or_mesh

        try:
            from_index = self.tags_to_indices[from_tag]
            to_index = self.tags_to_indices[to_tag]
        except KeyError as e:
            raise KeyError("No such tag: %s" % e.message)

        if from_index == to_index:
            result_points = points
        elif from_index < to_index:
            from_range = from_index, to_index
            result_points = self._transform(points, from_range=from_range)
        else:
            from_range = to_index, from_index
            result_points = self._transform(points, from_range=from_range, reverse=True)

        if hasattr(points_or_mesh, "v"):
            # for lace or those object with copy method, invoke its own copy method
            # otherwise just shallow copy
            result_mesh = (
                points_or_mesh.copy()
                if hasattr(points_or_mesh, "copy")
                else copy(points_or_mesh)
            )
            result_mesh.v = result_points

            return result_mesh
        else:
            return result_points

    def __setattr__(self, name, value):
        """
        value: Either an nx3 array of points or an instance of Mesh.

        """
        if name not in self.tags_to_indices:
            raise AttributeError("No such tag: %s" % name)

        self.__dict__["points_tag"] = name
        self.__dict__["points_or_mesh"] = value

    def __getattr__(self, name):
        if self.points_tag is None:
            raise ValueError("Must set a value before trying to read one")

        return self.do_transform(
            points_or_mesh=self.points_or_mesh, from_tag=self.points_tag, to_tag=name
        )



import unittest


class TestRounding(unittest.TestCase):
    def test_round_to(self):
        from .rounding import round_to

        self.assertEqual(round_to(3.8721, 0.05), 3.85)
        self.assertEqual(round_to(3.8721, 0.1), 3.9)
        self.assertEqual(round_to(3.8721, 0.25), 3.75)
        self.assertEqual(round_to(3.8721, 2.0), 4)
        self.assertEqual(round_to(3.8721, 2), 4)
        self.assertEqual(round_to(4, 0.33), 3.96)



class setter_property(object):
    """
    Note: setter_property incorrectly triggers method-hidden in pylint.

    """

    def __init__(self, fn, doc=None):
        self.fn = fn
        self.__name__ = fn.__name__
        self.__doc__ = doc if doc is not None else fn.__doc__

    def __get__(self, obj, cls):
        if self.__name__ in obj.__dict__:
            return obj.__dict__[self.__name__]
        else:
            return None

    def __set__(self, obj, value):
        return self.fn(obj, value)

    def __delete__(self, obj):
        return self.fn(obj, None)



import unittest
import numpy as np
from .predicates import is_empty_arraylike


class TestPredicates(unittest.TestCase):
    def test_predicates(self):
        from .predicates import isnumeric, isnumericarray

        self.assertTrue(isnumeric(42))
        self.assertTrue(isnumeric(42.0))
        self.assertFalse(isnumeric("42"))

        self.assertTrue(isnumericarray(np.array([1, 2, 3])))
        self.assertTrue(isnumericarray([1, 2, 3]))
        self.assertTrue(isnumericarray([1, 2.0, 3]))
        self.assertFalse(isnumericarray([1, 2.0, "3"]))
        self.assertFalse(isnumericarray(42))

    def test_predicates_some_more(self):
        import decimal
        from .predicates import isnumeric, isnumericarray, isnumericscalar

        def assertNumericArray(val):
            self.assertTrue(isnumeric(val))
            self.assertTrue(isnumericarray(val))
            self.assertFalse(isnumericscalar(val))

        def assertNumericScalar(val):
            self.assertTrue(isnumeric(val))
            self.assertFalse(isnumericarray(val))
            self.assertTrue(isnumericscalar(val))

        def assertNonNumeric(val):
            self.assertFalse(isnumeric(val))
            self.assertFalse(isnumericarray(val))
            self.assertFalse(isnumericscalar(val))

        assertNumericScalar(5)
        assertNumericScalar(123.456)
        assertNumericScalar(decimal.Decimal("123.456"))

        assertNumericArray([1, 2, 3])
        assertNumericArray(np.array([1, 2, 3]))
        a = np.array([1, 2.3, 4.5, 6.7, 8.9])
        assertNumericArray(a)
        assertNumericScalar(a[1])
        assertNumericArray(np.array([np.array([1]), np.array([2])]))

        assertNonNumeric(True)
        assertNonNumeric("123.345")
        assertNonNumeric(np.nan)
        assertNonNumeric(np.inf)
        assertNonNumeric(np.array(["1"]))
        assertNonNumeric(np.array(["foo", "bar", "baz"]))
        assertNonNumeric([1, "2", 3])
        assertNonNumeric(1 + 1j)
        assertNonNumeric(np.array([1 + 1j]))
        assertNonNumeric(1 + 0j)
        assertNonNumeric(np.array([1 + 0j]))
        assertNonNumeric([True, False, True])
        assertNonNumeric(np.array([True, False, True]))

    def test_is_empty_arraylike_none(self):
        array = None
        self.assertTrue(is_empty_arraylike(array))

    def test_is_empty_arraylike_list(self):
        array = []
        self.assertTrue(is_empty_arraylike(array))

    def test_is_empty_arraylike_non_empty_list(self):
        array = [1]
        self.assertFalse(is_empty_arraylike(array))

    def test_is_empty_arraylike_tuple(self):
        array = ()
        self.assertTrue(is_empty_arraylike(array))

    def test_is_empty_arraylike_non_empty_tuple(self):
        array = (1,)
        self.assertFalse(is_empty_arraylike(array))

    def test_is_empty_arraylike_ndarray(self):
        array = np.array([])
        self.assertTrue(is_empty_arraylike(array))

    def test_is_empty_arraylike_non_empty_ndarray(self):
        array = np.array([1])
        self.assertFalse(is_empty_arraylike(array))



import unittest
from . import units


class TestUnits(unittest.TestCase):
    def test_all_units(self):
        for x in ["in", "cm", "deg", "years"]:
            self.assertIn(x, units.all_units)
        for method in ["all_units", "factor", "convert"]:
            self.assertNotIn(method, units.all_units)

    def test_all_units_classes(self):
        self.assertEqual(
            set(units.all_units_classes),
            set(["length", "weight", "angle", "time", "time_rate"]),
        )

    def test_units_class(self):
        self.assertEqual(units.units_class("cm"), "length")
        self.assertEqual(units.units_class("stone"), "weight")
        self.assertEqual(units.units_class("rad"), "angle")
        self.assertEqual(units.units_class("sec"), "time")
        self.assertEqual(units.units_class(""), None)
        self.assertEqual(units.units_class(None), None)

    def test_units_in_class(self):
        self.assertIn("m", units.units_in_class("length"))

    def test_lengths(self):
        self.assertIn("m", units.lengths)

    def test_weights(self):
        self.assertIn("lb", units.weights)

    def test_angles(self):
        self.assertIn("degrees", units.angles)

    def test_times(self):
        self.assertIn("sec", units.times)

    def test_time_rates(self):
        self.assertIn("hours_per_week", units.time_rates)

    def test_default_units(self):
        self.assertEqual(
            units.default_units(),
            {
                "length": "cm",
                "weight": "kg",
                "angle": "deg",
                "time": "yr",
                "time_rate": "hours_per_week",
            },
        )

    def test_raw(self):
        self.assertEqual(units.raw("m"), 1.0)
        self.assertEqual(units.raw("mm"), 0.001)

    def test_factor(self):
        with self.assertRaisesRegexp(
            ValueError,
            r"Can\'t convert between apples and oranges \(stone and fathoms\)",
        ):
            units.factor("stone", "fathoms")

        nones = [(None, ""), ("", None), (None, None), ("", "")]
        for value in nones:
            self.assertEqual(units.factor(*value), 1.0)

        self.assertEqual(units.factor("in", "cm"), 2.54)
        self.assertAlmostEqual(units.factor("year", "min"), 525600, -3)
        self.assertEqual(units.factor("sec", "hr", units_class="time"), 1.0 / 3600)

        with self.assertRaisesRegexp(
            ValueError, r"Units class must be length, but got time"
        ):
            self.assertEqual(
                units.factor("sec", "hr", units_class="length"), 1.0 / 3600
            )

    def test_convert(self):
        import numpy as np

        self.assertEqual(units.convert(25, "cm", "mm")[0], 250)
        self.assertEqual(units.convert(180, "deg", "rad")[0], np.pi)
        self.assertAlmostEqual(units.convert(-10, "ft", "in")[0], -120)

        self.assertEqual(units.convert(25, "cm", "mm")[1], "mm")
        self.assertEqual(units.convert(180, "deg", "rad")[1], "rad")
        self.assertAlmostEqual(units.convert(-10, "ft", "in")[1], "in")

    def test_conversion_factors(self):
        import numpy as np

        self.assertAlmostEqual(units.convert(25, "cm", "m")[0], 0.25)
        self.assertAlmostEqual(units.convert(25, "cm", "cm")[0], 25)
        self.assertAlmostEqual(units.convert(25, "cm", "mm")[0], 250)
        self.assertAlmostEqual(
            units.convert(25, "cm", "in")[0], 9.8425197
        )  # from google
        self.assertAlmostEqual(units.convert(25, "cm", "ft")[0], 0.82021)  # from google
        self.assertAlmostEqual(
            units.convert(25, "cm", "fathoms")[0], 0.136701662
        )  # from google
        self.assertAlmostEqual(
            units.convert(25, "cm", "cubits")[0], 0.546806649
        )  # from google
        self.assertAlmostEqual(units.convert(10, "kg", "kg")[0], 10)
        self.assertAlmostEqual(units.convert(1, "kg", "g")[0], 1000)
        self.assertAlmostEqual(
            units.convert(10, "kg", "lbs")[0], 22.0462
        )  # from google
        self.assertAlmostEqual(
            units.convert(10, "kg", "stone")[0], 1.57473
        )  # from google
        self.assertAlmostEqual(units.convert(90, "deg", "rad")[0], np.pi / 2)
        self.assertAlmostEqual(units.convert(90, "deg", "deg")[0], 90)
        self.assertAlmostEqual(units.convert(30, "min", "sec")[0], 30 * 60)
        self.assertAlmostEqual(units.convert(30, "min", "minutes")[0], 30)
        self.assertAlmostEqual(units.convert(30, "min", "hours")[0], 0.5)
        self.assertAlmostEqual(units.convert(2, "days", "min")[0], 2 * 24 * 60)
        self.assertAlmostEqual(units.convert(1, "years", "min")[0], 525948.48)

    def test_convert_list(self):
        self.assertEqual(units.convert_list([10, 20, 30], "cm", "mm"), [100, 200, 300])

    def test_convert_to_system_default(self):
        self.assertEqual(units.convert_to_system_default(10, "mm"), (1, "cm"))

        result = (
            units.convert_to_system_default(10, "mm", "united_states"),
            (0.393701, "in"),
        )
        self.assertAlmostEqual(result[0][0], result[1][0], 6)
        self.assertEqual(result[0][1], result[1][1])

        result = (
            units.convert_to_system_default(10, "stone", "united_states"),
            (140, "lb"),
        )
        self.assertAlmostEqual(result[0][0], result[1][0], 0)
        self.assertEqual(result[0][1], result[1][1])

        self.assertEqual(units.convert_to_system_default(10, ""), (10, ""))

    def test_prettify(self):
        self.assertEqual(units.prettify(182.13992, "cm"), (182.0, "cm", 71.75, "in"))
        self.assertEqual(units.prettify(1821.3992, "mm"), (182.0, "cm", 71.75, "in"))
        self.assertEqual(units.prettify(-182.13992, "cm"), (-182.0, "cm", -71.75, "in"))
        self.assertEqual(
            units.prettify(182.13992, "cm", precision=2), (182.14, "cm", 71.71, "in")
        )
        self.assertEqual(
            units.prettify(182.13992, "cm", precision=4),
            (182.1399, "cm", 71.7086, "in"),
        )



import unittest
import numpy as np
from .composite import CompositeTransform


def create_cube_verts(origin, size):
    # Create a cube. Since CompositeTransform just works on verticies,
    # we don't need a full lace.mesh object.
    origin = np.asarray(origin)
    size = np.repeat(size, 3)
    lower_base_plane = np.array(
        [
            # Lower base plane
            origin,
            origin + np.array([size[0], 0, 0]),
            origin + np.array([size[0], 0, size[2]]),
            origin + np.array([0, 0, size[2]]),
        ]
    )
    upper_base_plane = lower_base_plane + np.array([0, size[1], 0])
    return np.vstack([lower_base_plane, upper_base_plane])


class TestCompositeTransform(unittest.TestCase):
    def setUp(self):
        self.cube_v = create_cube_verts([1.0, 0.0, 0.0], 4.0)

    def test_translate(self):
        transform = CompositeTransform()
        transform.translate(np.array([8.0, 6.0, 7.0]))

        # Sanity checking.
        np.testing.assert_array_equal(self.cube_v[0], [1.0, 0.0, 0.0])
        np.testing.assert_array_equal(self.cube_v[6], [5.0, 4.0, 4.0])

        self.cube_v = transform(self.cube_v)

        np.testing.assert_array_equal(self.cube_v[0], [9.0, 6.0, 7.0])
        np.testing.assert_array_equal(self.cube_v[6], [13.0, 10.0, 11.0])

    def test_translate_by_list(self):
        transform = CompositeTransform()
        transform.translate([8.0, 6.0, 7.0])

        # Sanity checking.
        np.testing.assert_array_equal(self.cube_v[0], [1.0, 0.0, 0.0])
        np.testing.assert_array_equal(self.cube_v[6], [5.0, 4.0, 4.0])

        self.cube_v = transform(self.cube_v)

        np.testing.assert_array_equal(self.cube_v[0], [9.0, 6.0, 7.0])
        np.testing.assert_array_equal(self.cube_v[6], [13.0, 10.0, 11.0])

    def test_scale(self):
        transform = CompositeTransform()
        transform.scale(10.0)

        # Sanity checking.
        np.testing.assert_array_equal(self.cube_v[0], [1.0, 0.0, 0.0])
        np.testing.assert_array_equal(self.cube_v[6], [5.0, 4.0, 4.0])

        self.cube_v = transform(self.cube_v)

        np.testing.assert_array_equal(self.cube_v[0], [10.0, 0.0, 0.0])
        np.testing.assert_array_equal(self.cube_v[6], [50.0, 40.0, 40.0])

    def test_translate_then_scale(self):
        transform = CompositeTransform()
        transform.translate(np.array([8.0, 6.0, 7.0]))
        transform.scale(10.0)

        # Sanity checking.
        np.testing.assert_array_equal(self.cube_v[0], [1.0, 0.0, 0.0])
        np.testing.assert_array_equal(self.cube_v[6], [5.0, 4.0, 4.0])

        self.cube_v = transform(self.cube_v)

        np.testing.assert_array_equal(self.cube_v[0], [90.0, 60.0, 70.0])
        np.testing.assert_array_equal(self.cube_v[6], [130.0, 100.0, 110.0])

    def test_scale_then_translate(self):
        transform = CompositeTransform()
        transform.scale(10.0)
        transform.translate(np.array([8.0, 6.0, 7.0]))

        # Sanity checking.
        np.testing.assert_array_equal(self.cube_v[0], [1.0, 0.0, 0.0])
        np.testing.assert_array_equal(self.cube_v[6], [5.0, 4.0, 4.0])

        self.cube_v = transform(self.cube_v)

        np.testing.assert_array_equal(self.cube_v[0], [18.0, 6.0, 7.0])
        np.testing.assert_array_equal(self.cube_v[6], [58.0, 46.0, 47.0])

    def test_rotate_then_translate(self):
        transform = CompositeTransform()
        transform.rotate(np.array([1.0, 2.0, 3.0]))
        transform.translate(np.array([3.0, 2.0, 1.0]))

        v = np.array([1.0, 0.0, 0.0]).reshape(-1, 3)

        # Forward.
        np.testing.assert_allclose(
            np.array([2.30507944, 1.80799303, 1.69297817]).reshape(-1, 3), transform(v)
        )
        # Reverse.
        np.testing.assert_allclose(
            np.array([1.08087689, -1.45082159, -2.3930779]).reshape(-1, 3),
            transform(v, reverse=True),
        )

    def test_reorient(self):
        # TODO We should also test a non-axis-aligned up and look.

        transform = CompositeTransform()
        transform.reorient(
            up=np.array([0.0, 1.0, 0.0]), look=np.array([-1.0, 0.0, 0.0])
        )

        # Sanity checking.
        np.testing.assert_array_equal(self.cube_v[0], [1.0, 0.0, 0.0])
        np.testing.assert_array_equal(self.cube_v[6], [5.0, 4.0, 4.0])

        self.cube_v = transform(self.cube_v)

        np.testing.assert_array_equal(self.cube_v[0], [0.0, 0.0, -1.0])
        np.testing.assert_array_equal(self.cube_v[6], [4, 4.0, -5.0])

    def test_rotate(self):
        ways_to_rotate_around_y_a_quarter_turn = [
            np.array([[0, 0, 1], [0, 1, 0], [-1, 0, 0]]),
            np.array([0, np.pi / 2, 0]),
            np.array([[0, np.pi / 2, 0]]),
            np.array([[0], [np.pi / 2], [0]]),
            [0, np.pi / 2, 0],
        ]
        for rot in ways_to_rotate_around_y_a_quarter_turn:
            transform = CompositeTransform()
            transform.rotate(rot)
            cube_v = self.cube_v.copy()

            # Sanity checking.
            np.testing.assert_array_equal(cube_v[0], [1.0, 0.0, 0.0])
            np.testing.assert_array_equal(cube_v[6], [5.0, 4.0, 4.0])

            cube_v = transform(cube_v)

            np.testing.assert_array_almost_equal(cube_v[0], [0.0, 0.0, -1.0])
            np.testing.assert_array_almost_equal(cube_v[6], [4, 4.0, -5.0])

    def test_reverse_transforms(self):
        transforms = [CompositeTransform() for _ in range(5)]

        transforms[1].translate(np.array([8.0, 6.0, 7.0]))

        transforms[2].scale(10.0)

        transforms[3].translate(np.array([8.0, 6.0, 7.0]))
        transforms[3].scale(10.0)

        transforms[4].scale(10.0)
        transforms[4].translate(np.array([8.0, 6.0, 7.0]))

        for transform in transforms:
            # Sanity checking.
            np.testing.assert_array_equal(self.cube_v[0], [1.0, 0.0, 0.0])
            np.testing.assert_array_equal(self.cube_v[6], [5.0, 4.0, 4.0])

            transformed = transform(self.cube_v)

            untransformed_v = transform(transformed, reverse=True)

            np.testing.assert_array_almost_equal(untransformed_v[0], [1.0, 0.0, 0.0])
            np.testing.assert_array_almost_equal(untransformed_v[6], [5.0, 4.0, 4.0])

    def test_forward_reverse_equivalence(self):
        transform = CompositeTransform()
        transform.rotate(np.array([1.0, 2.0, 3.0]))
        transform.translate(np.array([3.0, 2.0, 1.0]))
        transform.scale(10.0)
        transform.rotate(np.array([7.0, 13.0, 5.0]))

        forward = transform.matrix_for()
        reverse = transform.matrix_for(reverse=True)
        np.testing.assert_allclose(reverse, np.linalg.inv(forward))

        forward = transform.matrix_for(from_range=(0, 2))
        reverse = transform.matrix_for(from_range=(0, 2), reverse=True)
        np.testing.assert_allclose(reverse, np.linalg.inv(forward))



import numpy as np
import vg


def surface_normal(points, normalize=True):
    """
    Compute the surface normal of a triangle.

    Can provide three points or an array of points.
    """
    p1 = points[..., 0, :]
    p2 = points[..., 1, :]
    p3 = points[..., 2, :]
    normal = np.cross(p2 - p1, p3 - p1)
    return vg.normalize(normal) if normalize else normal



def find_rigid_transform(a, b, visualize=False):
    """
    Args:
        a: a 3xN array of vertex locations
        b: a 3xN array of vertex locations

    Returns: (R,T) such that R.dot(a)+T ~= b
    Based on Arun et al, "Least-squares fitting of two 3-D point sets," 1987.
    See also Eggert et al, "Estimating 3-D rigid body transformations: a
    comparison of four major algorithms," 1997.
    """
    import numpy as np
    import scipy.linalg
    from .._temporary.matlab import col

    if a.shape[0] != 3:
        if a.shape[1] == 3:
            a = a.T
    if b.shape[0] != 3:
        if b.shape[1] == 3:
            b = b.T
    assert a.shape[0] == 3
    assert b.shape[0] == 3

    a_mean = np.mean(a, axis=1)
    b_mean = np.mean(b, axis=1)
    a_centered = a - col(a_mean)
    b_centered = b - col(b_mean)

    c = a_centered.dot(b_centered.T)
    u, s, v = np.linalg.svd(c, full_matrices=False)
    v = v.T
    R = v.dot(u.T)

    if scipy.linalg.det(R) < 0:
        if np.any(s == 0):  # This is only valid in the noiseless case; see the paper
            v[:, 2] = -v[:, 2]
            R = v.dot(u.T)
        else:
            raise ValueError(
                "find_rigid_transform found a reflection that it cannot recover from. Try RANSAC or something..."
            )

    T = col(b_mean - R.dot(a_mean))

    if visualize != False:
        from lace.mesh import Mesh
        from lace.meshviewer import MeshViewer

        mv = MeshViewer() if visualize is True else visualize
        a_T = R.dot(a) + T
        mv.set_dynamic_meshes(
            [
                Mesh(v=a.T, f=[]).set_vertex_colors("red"),
                Mesh(v=b.T, f=[]).set_vertex_colors("green"),
                Mesh(v=a_T.T, f=[]).set_vertex_colors("orange"),
            ]
        )

    return R, T


def find_rigid_rotation(a, b, allow_scaling=False):
    """
    Args:
        a: a 3xN array of vertex locations
        b: a 3xN array of vertex locations

    Returns: R such that R.dot(a) ~= b

    See link: http://en.wikipedia.org/wiki/Orthogonal_Procrustes_problem
    """
    import numpy as np
    import scipy.linalg
    from blmath.numerics.matlab import col

    assert a.shape[0] == 3
    assert b.shape[0] == 3

    if a.size == 3:
        cx = np.cross(a.ravel(), b.ravel())
        a = np.hstack((col(a), col(cx)))
        b = np.hstack((col(b), col(cx)))

    c = a.dot(b.T)
    u, _, v = np.linalg.svd(c, full_matrices=False)
    v = v.T
    R = v.dot(u.T)

    if scipy.linalg.det(R) < 0:
        v[:, 2] = -v[:, 2]
        R = v.dot(u.T)

    if allow_scaling:
        scalefactor = scipy.linalg.norm(b) / scipy.linalg.norm(a)
        R = R * scalefactor

    return R



import unittest
import numpy as np
from .translation import translation


class TranslationsTest(unittest.TestCase):
    def setUp(self):
        self.translation = translation

        # test data: square of side 1 with lower left corner at (1, 1)
        self.points = np.array([[1.0, 1.0], [1.0, 2.0], [2.0, 2.0], [2.0, 1.0]])

    def test_translates_and_returns_translation_factor_with_translation_factor(self):
        points = self.points
        translation_factor = np.array([-1.0, 2.0])

        expected_image = np.array([[0.0, 3.0], [0.0, 4.0], [1.0, 4.0], [1.0, 3.0]])

        image, returned_translation_factor = translation(points, translation_factor)

        np.testing.assert_array_equal(image, expected_image)
        np.testing.assert_array_equal(returned_translation_factor, translation_factor)

    def test_translates_with_center_of_mass_and_returns_translation_factor_with_no_translation_factor(
        self
    ):
        points = self.points
        expected_image = np.array([[-0.5, -0.5], [-0.5, 0.5], [0.5, 0.5], [0.5, -0.5]])
        expected_translation_factor = -np.mean(points, axis=0)

        image, translation_factor = translation(points)

        np.testing.assert_array_equal(image, expected_image)
        np.testing.assert_array_equal(translation_factor, expected_translation_factor)



def round_to(value, nearest):
    """
    Round a number to the nearest specified increment.

    e.g.:

        >>> round_to(3.8721, 0.05)
        3.85
        >>> round_to(3.8721, 0.1)
        3.9
        >>> round_to(3.8725, 0.25)
        3.75
        >>> round_to(3.8725, 2.0)
        4

    Use reciprocal due to floating point weirdness:

        >>> value, nearest = 3.9, 0.1
        >>> round(value / nearest) * nearest
        3.9000000000000004
        >>> round(value / nearest) / (1.0 / nearest)
        3.9

    """
    reciprocal = 1.0 / nearest
    return round(value / nearest) / reciprocal


def rounded_list(iterable, precision=0):
    return [(None if x is None else round(x, precision)) for x in iterable]



import numpy as np


def translation(points, translation_factor=None):
    """
    Args:

        - points:
            an np.ndarray of 1 x N vectors
        - translation_factor (optional, default is the mean vector of 'points'):
            a 1 x N np.ndarray

    Returns a tuple (image, translation_factor):

        - image:
                the image of 'points' under the translation, with its order preserved.
        - translation_factor:
                the translation_factor. this is useful when translation_factor
                is initially None, and the caller wants to apply the inverse
    """
    if translation_factor is None:
        translation_factor = -np.mean(points, axis=0)

    image = points + translation_factor

    return (image, translation_factor)



def isnumeric(x):
    import numpy as np

    if hasattr(x, "__iter__"):
        x = np.asarray(x)
    if (
        isinstance(x, bool)
        or isinstance(x, np.bool)
        or (hasattr(x, "dtype") and x.dtype == np.bool)
    ):
        # Special case; bools do support math, but don't match our definition of numeric
        return False
    if not np.all(np.isrealobj(x)):
        # We don't want complex numbers
        return False
    try:
        if np.any(np.isnan(x)) or np.any(np.isinf(x)):
            return False
    except TypeError:  # isnan is not supported on some builtin numeric types like Decimal
        pass
    try:
        # See if it supports mathiness
        import warnings

        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            x + x, x - x, x * x, x / x  # pylint: disable=pointless-statement
    except ZeroDivisionError:
        return True
    except (TypeError, Exception):  # FIXME pylint: disable=broad-except
        return False
    else:
        return True


def isnumericscalar(x):
    return not hasattr(x, "__iter__") and isnumeric(x)


def isnumericarray(x):
    return hasattr(x, "__iter__") and isnumeric(x)


def is_empty_arraylike(arraylike):
    """Check if arraylike is None or empty.

    Arg:
        arraylike: None or np.ndarray or list or tuple

    Return:
        a boolean indicates if this array is empty
    """
    # pylint: disable=len-as-condition
    import numpy as np

    if arraylike is None:
        return True
    if isinstance(arraylike, np.ndarray) and arraylike.size == 0:
        return True
    if isinstance(arraylike, list) and len(arraylike) == 0:
        return True
    if isinstance(arraylike, tuple) and len(arraylike) == 0:
        return True
    return False



import unittest
import numpy as np
from .line_intersect import line_intersect, line_intersect3


class TestLineIntersect(unittest.TestCase):
    def test_line_intersect(self):
        p0, q0 = np.array([[0.0, 3.0], [4.0, 11.0]])
        p1, q1 = np.array([[-2.0, 8.0], [6.0, 4.0]])
        result = line_intersect(p0, q0, p1, q1)
        np.testing.assert_array_equal(result, [1.6, 6.2])

    def test_line_intersect_duplicate_point(self):
        p0, q0 = np.array([[0.0, 3.0], [5.0, 5.0]])
        p1, q1 = np.array([[5.0, 5.0], [6.0, 4.0]])
        result = line_intersect(p0, q0, p1, q1)
        np.testing.assert_array_equal(result, [5.0, 5.0])


class TestLineIntersect3D(unittest.TestCase):
    def test_line_intersect3_with_colinear_lines(self):
        p0, q0 = np.array([[0.0, 1.0, 2.0], [0.0, 10.0, 20.0]])
        p1, q1 = np.array([[0.0, 2.0, 4.0], [0.0, 4.0, 8.0]])
        result = line_intersect3(p0, q0, p1, q1)
        self.assertIsNone(result)

    def test_line_intersect3_with_parallel_lines(self):
        p0, q0 = np.array([[0.0, 1.0, 2.0], [0.0, 10.0, 20.0]])
        p1, q1 = np.array([[1.0, 2.0, 3.0], [1.0, 11.0, 21.0]])
        result = line_intersect3(p0, q0, p1, q1)
        self.assertIsNone(result)

    def test_line_intersect3_with_degenerate_input_p(self):
        p0, q0 = np.array([[0.0, 1.0, 2.0], [0.0, 10.0, 20.0]])
        p1, q1 = np.array([[0.0, 1.0, 2.0], [1.0, 11.0, 21.0]])
        result = line_intersect3(p0, q0, p1, q1)
        np.testing.assert_array_equal(result, [0.0, 1.0, 2.0])

    def test_line_intersect3_with_degenerate_input_q(self):
        p0, q0 = np.array([[0.0, 1.0, 2.0], [0.0, 10.0, 20.0]])
        p1, q1 = np.array([[1.0, 2.0, 3.0], [0.0, 10.0, 20.0]])
        result = line_intersect3(p0, q0, p1, q1)
        np.testing.assert_array_equal(result, [0.0, 10.0, 20.0])

    def test_line_intersect3_example_1(self):
        # This example tests the codirectional cross product case
        p0, q0 = np.array([[5.0, 5.0, 4.0], [10.0, 10.0, 6.0]])
        p1, q1 = np.array([[5.0, 5.0, 5.0], [10.0, 10.0, 3.0]])
        result = line_intersect3(p0, q0, p1, q1)
        np.testing.assert_array_equal(result, [25.0 / 4, 25.0 / 4, 9.0 / 2])

    def test_line_intersect3_example_2(self):
        # This example tests the opposite direction cross product case
        p0, q0 = np.array([[5.0, 5.0, 4.0], [10.0, 10.0, -6.0]])
        p1, q1 = np.array([[5.0, 5.0, 5.0], [10.0, 10.0, -3.0]])
        result = line_intersect3(p0, q0, p1, q1)
        np.testing.assert_array_equal(result, [2.5, 2.5, 9])

    def test_line_intersect3_example_3(self):
        p0, q0 = np.array([[6.0, 8.0, 4.0], [12.0, 15.0, 4.0]])
        p1, q1 = np.array([[6.0, 8.0, 2.0], [12.0, 15.0, 6.0]])
        result = line_intersect3(p0, q0, p1, q1)
        np.testing.assert_array_equal(result, [9.0, 23.0 / 2, 4.0])



import numpy as np


def partition(v, partition_size=5):
    """

    params:
        v:
            V x N np.array of points in N-space

        partition_size:
            how many partitions intervals for each segment?

    Fill in the line segments determined by v with equally
    spaced points - the space for each segment is determined
    by the length of the segment and the supplied partition size.

    """
    src = np.arange(len(v) - 1)
    dst = src + 1

    diffs = v[dst] - v[src]

    sqdis = np.square(diffs)
    dists = np.sqrt(np.sum(sqdis, axis=1))

    unitds = diffs / dists[:, np.newaxis]
    widths = dists / partition_size

    domain = widths[:, np.newaxis] * np.arange(0, partition_size)
    domain = domain.flatten()[:, np.newaxis]

    points = np.repeat(v[:-1], partition_size, axis=0)
    unitds = np.repeat(unitds, partition_size, axis=0)

    filled = points + (unitds * domain)

    return np.vstack((filled, v[-1]))


def partition_segment(p1, p2, n_samples, endpoint=True):
    """
    For two points in n-space, return an np.ndarray of equidistant partition
    points along the segment determined by p1 & p2.

    The total number of points returned will be n_samples. When n_samples is
    2, returns the original points.

    When endpoint is True, p2 is the last point. When false, p2 is excluded.

    Partition order is oriented from p1 to p2.

    Args:
        p1, p2:
            1 x N vectors

        partition_size:
            size of partition. should be >= 2.

    """
    if not isinstance(n_samples, int):
        raise TypeError("partition_size should be an int.")
    elif n_samples < 2:
        raise ValueError("partition_size should be bigger than 1.")

    return (p2 - p1) * np.linspace(0, 1, num=n_samples, endpoint=endpoint)[
        :, np.newaxis
    ] + p1


def partition_segment_old(p1, p2, partition_size=5):
    """
    Deprecated. Please use partition_segment.

    For two points in n-space, return an np.ndarray of partition points at equal widths
    determined by 'partition_size' on the interior of the segment determined by p1 & p2.

    Accomplished by partitioning the segment into 'partition_size' sub-intervals.

    Partition order is oriented from p1 to p2.

    Args:
        p1, p2:
            1 x N vectors

        partition_size:
            size of partition. should be > 1.
    """

    if not isinstance(partition_size, int):
        raise TypeError("partition_size should be an int.")
    elif partition_size < 2:
        raise ValueError("partition_size should be bigger than 1.")

    dist = np.linalg.norm(p1 - p2)

    unit_direction = (p2 - p1) / dist
    partition_width = dist / partition_size

    domain = partition_width * np.arange(1, partition_size)

    return p1 + unit_direction * domain[:, np.newaxis]



from .composite import CompositeTransform
from .coordinate_manager import CoordinateManager
from .rigid_transform import find_rigid_transform
from .rigid_transform import find_rigid_rotation
from .rotation import rotation_from_up_and_look



import unittest
import numpy as np
from .rotation import rotation_from_up_and_look, euler


class TestRotationFromUpAndLook(unittest.TestCase):
    def test_starting_with_canonical_reference_frame_gives_identity(self):
        result = rotation_from_up_and_look(up=[0, 1, 0], look=[0, 0, 1])
        np.testing.assert_array_almost_equal(result, np.eye(3))

    def test_raises_value_error_with_zero_length_inputs(self):
        with self.assertRaises(ValueError):
            rotation_from_up_and_look(up=[0, 0, 0], look=[0, 0, 1])
        with self.assertRaises(ValueError):
            rotation_from_up_and_look(up=[0, 1, 0], look=[0, 0, 0])

    def test_raises_value_error_with_colinear_inputs(self):
        with self.assertRaises(ValueError):
            rotation_from_up_and_look(up=[0, 0, 1], look=[0, 0, 1])
        with self.assertRaises(ValueError):
            rotation_from_up_and_look(up=[0, 0, 1], look=[0, 0, -1])

    def test_normalizes_inputs(self):
        result = rotation_from_up_and_look(up=[0, 42, 0], look=[0, 0, 13])
        np.testing.assert_array_almost_equal(result, np.eye(3))

    def test_always_outputs_float64(self):
        result = rotation_from_up_and_look(
            up=np.array([0, 1, 0], dtype=np.float32),
            look=np.array([0, 0, 1], dtype=np.float32),
        )
        self.assertEqual(result.dtype, np.float64)
        np.testing.assert_array_almost_equal(result, np.eye(3))

    def test_vary_look_alone(self):
        np.testing.assert_array_almost_equal(
            euler([0, 45]), rotation_from_up_and_look(up=[0, 1, 0], look=[-1, 0, 1])
        )
        np.testing.assert_array_almost_equal(
            euler([0, 90]), rotation_from_up_and_look(up=[0, 1, 0], look=[-1, 0, 0])
        )
        np.testing.assert_array_almost_equal(
            euler([0, 135]), rotation_from_up_and_look(up=[0, 1, 0], look=[-1, 0, -1])
        )
        np.testing.assert_array_almost_equal(
            euler([0, 180]), rotation_from_up_and_look(up=[0, 1, 0], look=[0, 0, -1])
        )
        np.testing.assert_array_almost_equal(
            euler([0, 225]), rotation_from_up_and_look(up=[0, 1, 0], look=[1, 0, -1])
        )
        np.testing.assert_array_almost_equal(
            euler([0, 270]), rotation_from_up_and_look(up=[0, 1, 0], look=[1, 0, 0])
        )
        np.testing.assert_array_almost_equal(
            euler([0, 315]), rotation_from_up_and_look(up=[0, 1, 0], look=[1, 0, 1])
        )

    def test_vary_up_alone(self):
        np.testing.assert_array_almost_equal(
            euler([0, 0, 45]), rotation_from_up_and_look(up=[1, 1, 0], look=[0, 0, 1])
        )
        np.testing.assert_array_almost_equal(
            euler([0, 0, 90]), rotation_from_up_and_look(up=[1, 0, 0], look=[0, 0, 1])
        )
        np.testing.assert_array_almost_equal(
            euler([0, 0, 135]), rotation_from_up_and_look(up=[1, -1, 0], look=[0, 0, 1])
        )
        np.testing.assert_array_almost_equal(
            euler([0, 0, 180]), rotation_from_up_and_look(up=[0, -1, 0], look=[0, 0, 1])
        )
        np.testing.assert_array_almost_equal(
            euler([0, 0, 225]),
            rotation_from_up_and_look(up=[-1, -1, 0], look=[0, 0, 1]),
        )
        np.testing.assert_array_almost_equal(
            euler([0, 0, 270]), rotation_from_up_and_look(up=[-1, 0, 0], look=[0, 0, 1])
        )
        np.testing.assert_array_almost_equal(
            euler([0, 0, 315]), rotation_from_up_and_look(up=[-1, 1, 0], look=[0, 0, 1])
        )


class TestEuler(unittest.TestCase):
    """
    Look up rotation matrices on wikipedia and you too can work these out by hand...
    """

    def test_x(self):
        s2 = np.sqrt(2) / 2
        np.testing.assert_array_almost_equal(euler(0), np.eye(3))
        np.testing.assert_array_almost_equal(
            euler(45), np.array([[1, 0, 0], [0, s2, -s2], [0, s2, s2]])
        )
        np.testing.assert_array_almost_equal(
            euler(90), np.array([[1, 0, 0], [0, 0, -1], [0, 1, 0]])
        )
        np.testing.assert_array_almost_equal(
            euler(180), np.array([[1, 0, 0], [0, -1, -0], [0, 0, -1]])
        )
        np.testing.assert_array_almost_equal(
            euler(270), np.array([[1, 0, 0], [0, 0, 1], [0, -1, 0]])
        )
        np.testing.assert_array_almost_equal(euler(360), np.eye(3))
        np.testing.assert_array_almost_equal(euler(0 * np.pi, units="rad"), np.eye(3))
        np.testing.assert_array_almost_equal(
            euler(0.25 * np.pi, units="rad"),
            np.array([[1, 0, 0], [0, s2, -s2], [0, s2, s2]]),
        )
        np.testing.assert_array_almost_equal(
            euler(0.5 * np.pi, units="rad"),
            np.array([[1, 0, 0], [0, 0, -1], [0, 1, 0]]),
        )
        np.testing.assert_array_almost_equal(
            euler(1 * np.pi, units="rad"),
            np.array([[1, 0, 0], [0, -1, -0], [0, 0, -1]]),
        )
        np.testing.assert_array_almost_equal(
            euler(1.5 * np.pi, units="rad"),
            np.array([[1, 0, 0], [0, 0, 1], [0, -1, 0]]),
        )
        np.testing.assert_array_almost_equal(euler(2 * np.pi, units="rad"), np.eye(3))

    def test_y(self):
        s2 = np.sqrt(2) / 2
        np.testing.assert_array_almost_equal(euler([0, 0]), np.eye(3))
        np.testing.assert_array_almost_equal(
            euler([0, 45]), np.array([[s2, 0, s2], [0, 1, 0], [-s2, 0, s2]])
        )
        np.testing.assert_array_almost_equal(
            euler([0, 90]), np.array([[0, 0, 1], [0, 1, 0], [-1, 0, 0]])
        )
        np.testing.assert_array_almost_equal(
            euler([0, 180]), np.array([[-1, 0, 0], [0, 1, 0], [-0, 0, -1]])
        )
        np.testing.assert_array_almost_equal(
            euler([0, 270]), np.array([[0, 0, -1], [0, 1, 0], [1, 0, 0]])
        )
        np.testing.assert_array_almost_equal(euler([0, 360]), np.eye(3))
        np.testing.assert_array_almost_equal(
            euler([0, 0 * np.pi], units="rad"), np.eye(3)
        )
        np.testing.assert_array_almost_equal(
            euler([0, 0.25 * np.pi], units="rad"),
            np.array([[s2, 0, s2], [0, 1, 0], [-s2, 0, s2]]),
        )
        np.testing.assert_array_almost_equal(
            euler([0, 0.5 * np.pi], units="rad"),
            np.array([[0, 0, 1], [0, 1, 0], [-1, 0, 0]]),
        )
        np.testing.assert_array_almost_equal(
            euler([0, 1 * np.pi], units="rad"),
            np.array([[-1, 0, 0], [0, 1, 0], [-0, 0, -1]]),
        )
        np.testing.assert_array_almost_equal(
            euler([0, 1.5 * np.pi], units="rad"),
            np.array([[0, 0, -1], [0, 1, 0], [1, 0, 0]]),
        )
        np.testing.assert_array_almost_equal(
            euler([0, 2 * np.pi], units="rad"), np.eye(3)
        )

    def test_z(self):
        s2 = np.sqrt(2) / 2
        np.testing.assert_array_almost_equal(euler([0, 0, 0]), np.eye(3))
        np.testing.assert_array_almost_equal(
            euler([0, 0, 45]), np.array([[s2, -s2, 0], [s2, s2, 0], [0, 0, 1]])
        )
        np.testing.assert_array_almost_equal(
            euler([0, 0, 90]), np.array([[0, -1, 0], [1, 0, 0], [0, 0, 1]])
        )
        np.testing.assert_array_almost_equal(
            euler([0, 0, 180]), np.array([[-1, 0, 0], [0, -1, 0], [0, 0, 1]])
        )
        np.testing.assert_array_almost_equal(
            euler([0, 0, 270]), np.array([[0, 1, 0], [-1, 0, 0], [0, 0, 1]])
        )
        np.testing.assert_array_almost_equal(euler([0, 0, 360]), np.eye(3))
        np.testing.assert_array_almost_equal(
            euler([0, 0, 0 * np.pi], units="rad"), np.eye(3)
        )
        np.testing.assert_array_almost_equal(
            euler([0, 0, 0.25 * np.pi], units="rad"),
            np.array([[s2, -s2, 0], [s2, s2, 0], [0, 0, 1]]),
        )
        np.testing.assert_array_almost_equal(
            euler([0, 0, 0.5 * np.pi], units="rad"),
            np.array([[0, -1, 0], [1, 0, 0], [0, 0, 1]]),
        )
        np.testing.assert_array_almost_equal(
            euler([0, 0, 1 * np.pi], units="rad"),
            np.array([[-1, 0, 0], [0, -1, 0], [0, 0, 1]]),
        )
        np.testing.assert_array_almost_equal(
            euler([0, 0, 1.5 * np.pi], units="rad"),
            np.array([[0, 1, 0], [-1, 0, 0], [0, 0, 1]]),
        )
        np.testing.assert_array_almost_equal(
            euler([0, 0, 2 * np.pi], units="rad"), np.eye(3)
        )

    # Really should write some tests for composition...



import numpy as np


def barycentric_coordinates_of_projection(p, q, u, v):
    """ Given a point, gives projected coords of that point to a triangle
    in barycentric coordinates.

    See Heidrich, Computing the Barycentric Coordinates of a Projected Point, JGT 05
    at http://www.cs.ubc.ca/~heidrich/Papers/JGT.05.pdf

    Args:
        p: point to project
        q: a vertex of the triangle to project into
        u,v: edges of the the triangle such that it has vertices q, q+u, q+v

    Returns:
        b: barycentric coordinates of p's projection in triangle defined by q,u,v
            vectorized so p,q,u,v can all be 3xN
    """

    p = p.T
    q = q.T
    u = u.T
    v = v.T

    n = np.cross(u, v, axis=0)
    s = np.sum(n * n, axis=0)

    # If the triangle edges are collinear, cross-product is zero,
    # which makes "s" 0, which gives us divide by zero. So we
    # make the arbitrary choice to set s to epsv (=numpy.spacing(1)),
    # the closest thing to zero
    if np.isscalar(s):
        s = s if s else np.spacing(1)
    else:
        s[s == 0] = np.spacing(1)

    oneOver4ASquared = 1.0 / s
    w = p - q
    b2 = np.sum(np.cross(u, w, axis=0) * n, axis=0) * oneOver4ASquared
    b1 = np.sum(np.cross(w, v, axis=0) * n, axis=0) * oneOver4ASquared
    b = np.vstack((1 - b1 - b2, b1, b2))

    return b.T



import numpy as np


def inflection_points(points, axis, span):
    """
    Find the list of vertices that preceed inflection points in a curve. The curve is differentiated
    with respect to the coordinate system defined by axis and span.

    axis: A vector representing the vertical axis of the coordinate system.
    span: A vector representing the the horiztonal axis of the coordinate system.

    returns: a list of points in space corresponding to the vertices that
    immediately preceed inflection points in the curve
    """

    coords_on_span = points.dot(span)
    dx = np.gradient(coords_on_span)
    coords_on_axis = points.dot(axis)

    # Take the second order finite difference of the curve with respect to the
    # defined coordinate system
    finite_difference_2 = np.gradient(np.gradient(coords_on_axis, dx), dx)

    # Compare the product of all neighboring pairs of points in the second derivative
    # If a pair of points has a negative product, then the second derivative changes sign
    # at one of those points, signalling an inflection point
    is_inflection_point = [
        finite_difference_2[i] * finite_difference_2[i + 1] <= 0
        for i in range(len(finite_difference_2) - 1)
    ]

    inflection_point_indices = [i for i, b in enumerate(is_inflection_point) if b]

    if len(inflection_point_indices) == 0:  # pylint: disable=len-as-condition
        return []

    return points[inflection_point_indices]



# pylint: disable=invalid-unary-operand-type
import unittest
import numpy as np
from .test_composite import create_cube_verts
from .coordinate_manager import CoordinateManager


class TestCoordinateManager(unittest.TestCase):
    def test_coordinate_manager_forward(self):
        cube_v = create_cube_verts([1.0, 0.0, 0.0], 4.0)
        cube_floor_point = np.array([3.0, 0.0, 2.0])  # as lace.mesh.floor_point

        coordinate_manager = CoordinateManager()
        coordinate_manager.tag_as("source")
        coordinate_manager.translate(-cube_floor_point)
        coordinate_manager.scale(2)
        coordinate_manager.tag_as("floored_and_scaled")
        coordinate_manager.translate(np.array([0.0, -4.0, 0.0]))
        coordinate_manager.tag_as("centered_at_origin")

        coordinate_manager.source = cube_v

        floored_and_scaled_v = coordinate_manager.do_transform(
            cube_v, from_tag="source", to_tag="floored_and_scaled"
        )

        # Sanity check
        np.testing.assert_array_almost_equal(cube_v[0], [1.0, 0.0, 0.0])
        np.testing.assert_array_almost_equal(cube_v[6], [5.0, 4.0, 4.0])

        np.testing.assert_array_almost_equal(floored_and_scaled_v[0], [-4.0, 0.0, -4.0])
        np.testing.assert_array_almost_equal(floored_and_scaled_v[6], [4.0, 8.0, 4.0])

        centered_at_origin_v_1 = coordinate_manager.do_transform(
            cube_v, from_tag="source", to_tag="centered_at_origin"
        )
        centered_at_origin_v_2 = coordinate_manager.do_transform(
            floored_and_scaled_v,
            from_tag="floored_and_scaled",
            to_tag="centered_at_origin",
        )

        np.testing.assert_array_almost_equal(
            centered_at_origin_v_1[0], [-4.0, -4.0, -4.0]
        )
        np.testing.assert_array_almost_equal(centered_at_origin_v_1[6], [4.0, 4.0, 4.0])

        np.testing.assert_array_almost_equal(
            centered_at_origin_v_2[0], [-4.0, -4.0, -4.0]
        )
        np.testing.assert_array_almost_equal(centered_at_origin_v_2[6], [4.0, 4.0, 4.0])

        source_v_1 = coordinate_manager.do_transform(
            floored_and_scaled_v, from_tag="floored_and_scaled", to_tag="source"
        )
        source_v_2 = coordinate_manager.do_transform(
            centered_at_origin_v_1, from_tag="centered_at_origin", to_tag="source"
        )
        np.testing.assert_array_almost_equal(source_v_1, cube_v)
        np.testing.assert_array_almost_equal(source_v_2, cube_v)

    def test_coordinate_manager_forward_with_attrs(self):
        cube_v = create_cube_verts([1.0, 0.0, 0.0], 4.0)
        cube_floor_point = np.array([3.0, 0.0, 2.0])  # as lace.mesh.floor_point

        coordinate_manager = CoordinateManager()
        coordinate_manager.tag_as("source")
        coordinate_manager.translate(-cube_floor_point)
        coordinate_manager.scale(2)
        coordinate_manager.tag_as("floored_and_scaled")
        coordinate_manager.translate(np.array([0.0, -4.0, 0.0]))
        coordinate_manager.tag_as("centered_at_origin")

        coordinate_manager.source = cube_v

        # Sanity check
        np.testing.assert_array_almost_equal(cube_v[0], [1.0, 0.0, 0.0])
        np.testing.assert_array_almost_equal(cube_v[6], [5.0, 4.0, 4.0])

        floored_and_scaled_v = coordinate_manager.floored_and_scaled
        np.testing.assert_array_almost_equal(floored_and_scaled_v[0], [-4.0, 0.0, -4.0])
        np.testing.assert_array_almost_equal(floored_and_scaled_v[6], [4.0, 8.0, 4.0])

        centered_at_origin_v = coordinate_manager.centered_at_origin
        np.testing.assert_array_almost_equal(
            centered_at_origin_v[0], [-4.0, -4.0, -4.0]
        )
        np.testing.assert_array_almost_equal(centered_at_origin_v[6], [4.0, 4.0, 4.0])

        source_v = coordinate_manager.source
        np.testing.assert_array_almost_equal(source_v, cube_v)

    def test_coordinate_manager_forward_on_mesh(self):
        import sys

        if sys.version_info >= (3, 3):
            from unittest.mock import MagicMock
        else:
            from mock import MagicMock

        cube_v = create_cube_verts([1.0, 0.0, 0.0], 4.0)
        cube_floor_point = np.array([3.0, 0.0, 2.0])  # as lace.mesh.floor_point

        # By default a magic mock will always have any attribute it's asked for;
        # here we set the spec property so that it will not respond to having a copy method
        # when the CoodinateManager looks for it.
        cube = MagicMock(
            spec=["v", "other_thing"], v=cube_v, other_thing=np.array([-9.0])
        )

        coordinate_manager = CoordinateManager()
        coordinate_manager.tag_as("source")
        coordinate_manager.translate(-cube_floor_point)
        coordinate_manager.scale(2)
        coordinate_manager.tag_as("floored_and_scaled")
        coordinate_manager.translate(np.array([0.0, -4.0, 0.0]))
        coordinate_manager.tag_as("centered_at_origin")

        coordinate_manager.source = cube

        # Sanity check
        np.testing.assert_array_almost_equal(cube.v[0], [1.0, 0.0, 0.0])
        np.testing.assert_array_almost_equal(cube.v[6], [5.0, 4.0, 4.0])
        np.testing.assert_array_equal(cube.other_thing, [-9.0])

        floored_and_scaled = coordinate_manager.floored_and_scaled
        np.testing.assert_array_almost_equal(floored_and_scaled.v[0], [-4.0, 0.0, -4.0])
        np.testing.assert_array_almost_equal(floored_and_scaled.v[6], [4.0, 8.0, 4.0])
        np.testing.assert_array_equal(floored_and_scaled.other_thing, [-9.0])



class Units(object):
    def __init__(self):
        import numpy as np

        self._units = units = {}

        # lengths: conversion to m
        units["m"] = units["meter"] = units["meters"] = ("length", 1.0)
        units["mm"] = units["millimeter"] = units["millimeters"] = (
            "length",
            units["m"][1] / 1000,
        )
        units["cm"] = units["centimeter"] = units["centimeters"] = (
            "length",
            units["m"][1] / 100,
        )
        units["in"] = units["inch"] = units["inches"] = ("length", 0.0254)
        units["ft"] = units["foot"] = units["feet"] = ("length", 0.3048)
        units["fathoms"] = ("length", 1.8288)
        units["cubits"] = ("length", 0.4572)

        # weights: conversion to kg
        units["kg"] = units["kilograms"] = ("weight", 1.0)
        units["g"] = units["grams"] = ("weight", units["kg"][1] / 1000)
        units["lbs"] = units["pounds"] = units["lb"] = (
            "weight",
            units["kg"][1] / 2.20462,
        )
        units["stone"] = ("weight", units["kg"][1] / 0.157473)

        # angles: conversion to degrees
        units["deg"] = units["degrees"] = ("angle", 1.0)
        units["rad"] = units["radians"] = ("angle", units["deg"][1] * 180 / np.pi)

        # time: conversion to seconds
        units["sec"] = units["second"] = units["seconds"] = ("time", 1.0)
        units["min"] = units["minute"] = units["minutes"] = (
            "time",
            units["sec"][1] * 60,
        )
        units["hr"] = units["hour"] = units["hours"] = ("time", 60 * units["min"][1])
        units["day"] = units["days"] = ("time", units["hour"][1] * 24)
        units["yr"] = units["year"] = units["years"] = (
            "time",
            units["day"][1] * 365.242,
        )

        units["hours_per_week"] = ("time_rate", 1.0)

        self._default_units = {
            "metric": {
                "length": "cm",
                "weight": "kg",
                "angle": "deg",
                "time": "yr",
                "time_rate": "hours_per_week",
            },
            "united_states": {
                "length": "in",
                "weight": "lb",
                "angle": "deg",
                "time": "yr",
                "time_rate": "hours_per_week",
            },
        }

    @property
    def all_units(self):
        """
        Return a list of all supported units.

        """
        return self._units.keys()

    @property
    def all_units_classes(self):
        """
        Return a list of all supported units classes, e.g.
        'length', 'weight', 'angle, 'time', 'time_rate'.

        """
        return list(set([item[0] for item in self._units.values()]))

    def units_class(self, units):
        """
        Returns 'length', 'weight', 'angle', 'time', or 'time_rate'.

            >>> units_class('cm')
            'length'

        """
        if units is None or units == "":
            return None
        return self._units[units][0]

    def units_in_class(self, uclass):
        """
        Return a list of all units in uclass, where uclass is e.g.
        length', 'weight', 'angle, 'time', 'time_rate'.
        """
        return [key for key, (uclass_0, _) in self._units.items() if uclass_0 == uclass]

    @property
    def lengths(self):
        """
        List of all supported lengths.
        """
        return self.units_in_class("length")

    @property
    def weights(self):
        """
        List of all supported weights.
        """
        return self.units_in_class("weight")

    @property
    def angles(self):
        """
        List of all supported angles.
        """
        return self.units_in_class("angle")

    @property
    def times(self):
        """
        List of all supported times.
        """
        return self.units_in_class("time")

    @property
    def time_rates(self):
        """
        List of all supported time rates.
        """
        return self.units_in_class("time_rate")

    def default_units(self, unit_system="metric", exceptions={}):
        """
        Get the default unit for a given unit system. unit_class is
        'length', 'weight', 'angle', or 'time'. unit_system is
        either 'metric' or 'united_states'.

        """
        return dict(
            list(self._default_units[unit_system].items()) + list(exceptions.items())
        )

    def raw(self, units):
        """
        Returns a raw units conversion factor. Try not to use this.
        Use factor() or convert() instead.

        """
        return self._units[units][1]

    def factor(
        self, from_units, to_units, units_class=None
    ):  # FIXME pylint: disable=redefined-outer-name
        """
        Return a conversion factor:

            >>> value_in_cm = 25
            >>> value_in_cm * factor('cm', 'mm')
            250

        class: If specified, the class of the units must match the class provided.

        """
        if (from_units is None or not len(from_units)) and (
            to_units is None or not len(to_units)
        ):  # pylint: disable=len-as-condition
            return 1.0
        if from_units == to_units:
            return 1.0
        if self._units[from_units][0] != self._units[to_units][0]:
            raise ValueError(
                "Can't convert between apples and oranges (%s and %s)"
                % (from_units, to_units)
            )
        if units_class and self._units[from_units][0] != units_class:
            raise ValueError(
                "Units class must be %s, but got %s"
                % (units_class, self._units[from_units][0])
            )
        return self._units[from_units][1] / self._units[to_units][1]

    def convert(
        self, value, from_units, to_units, units_class=None
    ):  # FIXME pylint: disable=redefined-outer-name
        """
        Convert a number from one unit to another.

        class: If specified, the class of the units must match the class provided.

        Returns a tuple with the converted value and the units.

            >>> value_cm = 25
            >>> value, units = convert(value_cm, 'cm', 'mm')
            >>> value
            250
            >>> units
            'mm'

        """
        # Get factor first so we return errors for apples and oranges,
        # even when value is None
        factor = self.factor(
            from_units, to_units, units_class=units_class
        )  # FIXME pylint: disable=redefined-outer-name
        if value is None:
            return None, None
        return value * factor, to_units

    def convert_list(self, a_list, from_units, to_units):
        """
        Convenience helper to convert a list of numbers from one unit to another.

        Unlike convert(), does not return a tuple.

            >>> convert_list([10, 20, 30], 'cm', 'mm')
            [100, 200, 300]

        """
        factor = self.factor(
            from_units, to_units
        )  # FIXME pylint: disable=redefined-outer-name
        return [factor * x for x in a_list]

    def convert_to_default(self, value, from_units, defaults):
        """
        Convert a number from the given unit to a default for
        the given unit system.

        Returns a tuple with the converted value and the units.

            >>> value_cm = 100
            >>> convert_to_default(value_cm, 'cm', {'length': 'in', 'weight': 'lb', 'angle': 'deg', 'time': 'yr'})
            (39.3701, 'in')

        """
        units_class = self.units_class(
            from_units
        )  # FIXME pylint: disable=redefined-outer-name
        if units_class:
            to_units = defaults[units_class]
            return self.convert(value, from_units, to_units)
        else:
            return value, from_units

    def convert_to_system_default(self, value, from_units, to_unit_system="metric"):
        """
        Convert a number from the given unit to a default for
        the given unit system.

        Returns a tuple with the converted value and the units.

            >>> value_cm = 100
            >>> convert_to_system_default(value_cm, 'cm', 'united_states')
            (39.3701, 'in')

        """
        return self.convert_to_default(
            value, from_units, self.default_units(to_unit_system)
        )

    def prettify(self, value, units, precision=None):
        """
        Take a value, and units, and return rounded values in the
        default metric and United States units.

        The default precision values vary per unit. Specifying an
        integer for `precision` will override those defaults.

            >>> prettify(182.13992, 'cm'),
            (182.0, 'cm', 71.75, 'in')
            >>> prettify(182.13992, 'cm', precision=1)
            (182.1, 'cm', 71.7, 'in')
            >>> prettify(182.13992, 'cm', precision=2)
            (182.14, 'cm', 71.71, 'in')

        """

        def round_value(value, units):
            from .rounding import round_to

            if precision is None:
                nearest = {"cm": 0.5, "in": 0.25, "kg": 0.5, "lb": 1.0, "yr": 1.0}.get(
                    units, 0.1
                )
            else:
                # Work around weird floating point rounding issues
                nearest = 1.0 / 10 ** precision
            return round_to(value, nearest), units

        return round_value(*self.convert_to_system_default(value, units)) + round_value(
            *self.convert_to_system_default(value, units, "united_states")
        )


_units = Units()

all_units = _units.all_units
all_units_classes = _units.all_units_classes
units_class = _units.units_class
units_in_class = _units.units_in_class
lengths = _units.lengths
weights = _units.weights
angles = _units.angles
times = _units.times
time_rates = _units.time_rates
default_units = _units.default_units
convert_to_default = _units.convert_to_default
raw = _units.raw
factor = _units.factor
convert = _units.convert
convert_list = _units.convert_list
convert_to_system_default = _units.convert_to_system_default
prettify = _units.prettify



import unittest
import numpy as np
import vg
from ..plane.plane import Plane
from .polyline import Polyline


class TestPolyline(unittest.TestCase):
    def test_edges(self):
        v = np.array(
            [
                [0.0, 0.0, 0.0],
                [1.0, 0.0, 0.0],
                [1.0, 1.0, 0.0],
                [1.0, 2.0, 0.0],
                [1.0, 3.0, 0.0],
            ]
        )

        expected_open = np.array([[0, 1], [1, 2], [2, 3], [3, 4]])

        np.testing.assert_array_equal(Polyline(v).e, expected_open)

        expected_closed = np.array([[0, 1], [1, 2], [2, 3], [3, 4], [4, 0]])

        np.testing.assert_array_equal(Polyline(v, closed=True).e, expected_closed)

    def test_segments(self):
        polyline = Polyline(
            np.array(
                [[0.0, 0.0, 0.0], [1.0, 0.0, 0.0], [1.0, 1.0, 0.0], [1.0, 2.0, 0.0]]
            ),
            closed=True,
        )

        expected = np.array(
            [
                [[0.0, 0.0, 0.0], [1.0, 0.0, 0.0]],
                [[1.0, 0.0, 0.0], [1.0, 1.0, 0.0]],
                [[1.0, 1.0, 0.0], [1.0, 2.0, 0.0]],
                [[1.0, 2.0, 0.0], [0.0, 0.0, 0.0]],
            ]
        )

        np.testing.assert_array_equal(polyline.segments, expected)

    def test_segment_vectors(self):
        polyline = Polyline(
            np.array(
                [[0.0, 0.0, 0.0], [1.0, 0.0, 0.0], [1.0, 1.0, 0.0], [1.0, 2.0, 0.0]]
            ),
            closed=True,
        )

        expected = np.array(
            [[1.0, 0.0, 0.0], [0.0, 1.0, 0.0], [0.0, 1.0, 0.0], [-1.0, -2.0, 0.0]]
        )

        np.testing.assert_array_equal(polyline.segment_vectors, expected)

    def test_length_of_empty_polyline(self):
        polyline = Polyline(None)
        self.assertEqual(polyline.total_length, 0)

        polyline = Polyline(None, closed=True)
        self.assertEqual(polyline.total_length, 0)

    def test_partition_by_length_noop(self):
        original = Polyline(
            np.array(
                [
                    [0.0, 0.0, 0.0],
                    [1.0, 0.0, 0.0],
                    [1.0, 1.0, 0.0],
                    [1.0, 2.0, 0.0],
                    [1.0, 3.0, 0.0],
                ]
            )
        )

        result = original.copy()
        indices = result.partition_by_length(1.0, ret_indices=True)

        expected_indices = np.array([0, 1, 2, 3, 4])

        np.testing.assert_array_almost_equal(result.v, original.v)
        np.testing.assert_array_equal(result.e, original.e)
        np.testing.assert_array_equal(indices, expected_indices)

    def test_partition_by_length_degenerate(self):
        """
        This covers a bug that arose from a numerical stability issue in
        measurement on EC2 / MKL.
        """
        original = Polyline(
            np.array([[0.0, 0.0, 0.0], [1.0, 0.0, 0.0], [1.0, 0.0, 0.0]])
        )

        result = original.copy()
        indices = result.partition_by_length(1.0, ret_indices=True)

        expected_indices = np.array([0, 1, 2])

        np.testing.assert_array_almost_equal(result.v, original.v)
        np.testing.assert_array_equal(result.e, original.e)
        np.testing.assert_array_equal(indices, expected_indices)

    def test_partition_by_length_divide_by_two(self):
        original = Polyline(
            np.array(
                [
                    [0.0, 0.0, 0.0],
                    [1.0, 0.0, 0.0],
                    [1.0, 1.0, 0.0],
                    [1.0, 2.0, 0.0],
                    [1.0, 3.0, 0.0],
                ]
            )
        )

        expected = Polyline(
            np.array(
                [
                    [0.0, 0.0, 0.0],
                    [0.5, 0.0, 0.0],
                    [1.0, 0.0, 0.0],
                    [1.0, 0.5, 0.0],
                    [1.0, 1.0, 0.0],
                    [1.0, 1.5, 0.0],
                    [1.0, 2.0, 0.0],
                    [1.0, 2.5, 0.0],
                    [1.0, 3.0, 0.0],
                ]
            )
        )

        expected_indices = np.array([0, 2, 4, 6, 8])

        for max_length in (0.99, 0.75, 0.5):
            result = original.copy()
            indices = result.partition_by_length(max_length, ret_indices=True)

            np.testing.assert_array_almost_equal(result.v, expected.v)
            np.testing.assert_array_equal(result.e, expected.e)
            np.testing.assert_array_equal(indices, expected_indices)

    def test_partition_length_divide_by_five(self):
        original = Polyline(
            np.array(
                [
                    [0.0, 0.0, 0.0],
                    [1.0, 0.0, 0.0],
                    [1.0, 1.0, 0.0],
                    [1.0, 2.0, 0.0],
                    [1.0, 3.0, 0.0],
                ]
            )
        )

        expected = Polyline(
            np.array(
                [
                    [0.0, 0.0, 0.0],
                    [0.2, 0.0, 0.0],
                    [0.4, 0.0, 0.0],
                    [0.6, 0.0, 0.0],
                    [0.8, 0.0, 0.0],
                    [1.0, 0.0, 0.0],
                    [1.0, 0.2, 0.0],
                    [1.0, 0.4, 0.0],
                    [1.0, 0.6, 0.0],
                    [1.0, 0.8, 0.0],
                    [1.0, 1.0, 0.0],
                    [1.0, 1.2, 0.0],
                    [1.0, 1.4, 0.0],
                    [1.0, 1.6, 0.0],
                    [1.0, 1.8, 0.0],
                    [1.0, 2.0, 0.0],
                    [1.0, 2.2, 0.0],
                    [1.0, 2.4, 0.0],
                    [1.0, 2.6, 0.0],
                    [1.0, 2.8, 0.0],
                    [1.0, 3.0, 0.0],
                ]
            )
        )

        expected_indices = np.array([0, 5, 10, 15, 20])

        for max_length in (0.2, 0.24):
            result = original.copy()
            indices = result.partition_by_length(max_length, ret_indices=True)

            np.testing.assert_array_almost_equal(result.v, expected.v)
            np.testing.assert_array_equal(result.e, expected.e)
            np.testing.assert_array_equal(indices, expected_indices)

    def test_partition_by_length_divide_some_leave_some(self):
        original = Polyline(
            np.array(
                [
                    [0.0, 0.0, 0.0],
                    [1.0, 0.0, 0.0],
                    [1.0, 1.0, 0.0],
                    [1.0, 7.0, 0.0],
                    [1.0, 8.0, 0.0],
                ]
            )
        )

        expected = Polyline(
            np.array(
                [
                    [0.0, 0.0, 0.0],
                    [1.0, 0.0, 0.0],
                    [1.0, 1.0, 0.0],
                    [1.0, 3.0, 0.0],
                    [1.0, 5.0, 0.0],
                    [1.0, 7.0, 0.0],
                    [1.0, 8.0, 0.0],
                ]
            )
        )

        expected_indices = np.array([0, 1, 2, 5, 6])

        for max_length in (2.0, 2.99):
            result = original.copy()
            indices = result.partition_by_length(max_length, ret_indices=True)

            np.testing.assert_array_almost_equal(result.v, expected.v)
            np.testing.assert_array_equal(result.e, expected.e)
            np.testing.assert_array_equal(indices, expected_indices)

    def test_partition_by_length_closed(self):
        original = Polyline(
            np.array(
                [
                    [0.0, 0.0, 0.0],
                    [1.0, 0.0, 0.0],
                    [1.0, 1.0, 0.0],
                    [1.0, 7.0, 0.0],
                    [1.0, 8.0, 0.0],
                    [0.0, 8.0, 0.0],
                ]
            ),
            closed=True,
        )

        expected = Polyline(
            np.array(
                [
                    [0.0, 0.0, 0.0],
                    [1.0, 0.0, 0.0],
                    [1.0, 1.0, 0.0],
                    [1.0, 3.0, 0.0],
                    [1.0, 5.0, 0.0],
                    [1.0, 7.0, 0.0],
                    [1.0, 8.0, 0.0],
                    [0.0, 8.0, 0.0],
                    [0.0, 6.0, 0.0],
                    [0.0, 4.0, 0.0],
                    [0.0, 2.0, 0.0],
                ]
            ),
            closed=True,
        )

        expected_indices = np.array([0, 1, 2, 5, 6, 7])

        for max_length in (2.0, 2.5, 2.6):
            result = original.copy()
            indices = result.partition_by_length(max_length, ret_indices=True)

            np.testing.assert_array_almost_equal(result.v, expected.v)
            np.testing.assert_array_equal(result.e, expected.e)
            np.testing.assert_array_equal(indices, expected_indices)

    def test_flip(self):
        original = Polyline(
            np.array(
                [
                    [0.0, 0.0, 0.0],
                    [1.0, 0.0, 0.0],
                    [1.0, 1.0, 0.0],
                    [1.0, 7.0, 0.0],
                    [1.0, 8.0, 0.0],
                    [0.0, 8.0, 0.0],
                ]
            ),
            closed=True,
        )

        expected = Polyline(
            np.array(
                [
                    [0.0, 8.0, 0.0],
                    [1.0, 8.0, 0.0],
                    [1.0, 7.0, 0.0],
                    [1.0, 1.0, 0.0],
                    [1.0, 0.0, 0.0],
                    [0.0, 0.0, 0.0],
                ]
            ),
            closed=True,
        )

        original.flip()

        np.testing.assert_array_almost_equal(original.v, expected.v)

    def test_reindexed(self):
        original = Polyline(
            np.array(
                [
                    [0.0, 0.0, 0.0],
                    [1.0, 0.0, 0.0],
                    [1.0, 1.0, 0.0],
                    [1.0, 7.0, 0.0],
                    [1.0, 8.0, 0.0],
                    [0.0, 8.0, 0.0],
                ]
            ),
            closed=True,
        )

        reindexed, edge_mapping = original.reindexed(5, ret_edge_mapping=True)

        expected = Polyline(
            np.array(
                [
                    [0.0, 8.0, 0.0],
                    [0.0, 0.0, 0.0],
                    [1.0, 0.0, 0.0],
                    [1.0, 1.0, 0.0],
                    [1.0, 7.0, 0.0],
                    [1.0, 8.0, 0.0],
                ]
            ),
            closed=True,
        )

        np.testing.assert_array_almost_equal(reindexed.v, expected.v)
        np.testing.assert_array_equal(
            original.segments[edge_mapping], reindexed.segments
        )

    def test_cut_by_plane_closed(self):
        original = Polyline(
            np.array(
                [
                    [0.0, 0.0, 0.0],
                    [1.0, 0.0, 0.0],
                    [1.0, 1.0, 0.0],
                    [1.0, 7.0, 0.0],
                    [1.0, 8.0, 0.0],
                    [0.0, 8.0, 0.0],
                ]
            ),
            closed=True,
        )

        expected = Polyline(
            np.array(
                [[1.0, 7.5, 0.0], [1.0, 8.0, 0.0], [0.0, 8.0, 0.0], [0.0, 7.5, 0.0]]
            ),
            closed=False,
        )
        actual = original.cut_by_plane(
            Plane(point_on_plane=np.array([0.0, 7.5, 0.0]), unit_normal=vg.basis.y)
        )

        np.testing.assert_array_almost_equal(actual.v, expected.v)
        self.assertFalse(actual.closed)

        expected = Polyline(
            np.array(
                [
                    [0.0, 7.5, 0.0],
                    [0.0, 0.0, 0.0],
                    [1.0, 0.0, 0.0],
                    [1.0, 1.0, 0.0],
                    [1.0, 7.0, 0.0],
                    [1.0, 7.5, 0.0],
                ]
            ),
            closed=False,
        )
        actual = original.cut_by_plane(
            Plane(point_on_plane=np.array([0.0, 7.5, 0.0]), unit_normal=vg.basis.neg_y)
        )

        np.testing.assert_array_almost_equal(actual.v, expected.v)
        self.assertFalse(actual.closed)

    def test_cut_by_plane_open(self):
        original = Polyline(
            np.array(
                [
                    [0.0, 0.0, 0.0],
                    [1.0, 0.0, 0.0],
                    [1.0, 1.0, 0.0],
                    [1.0, 7.0, 0.0],
                    [1.0, 8.0, 0.0],
                ]
            ),
            closed=False,
        )

        expected = Polyline(np.array([[1.0, 7.5, 0.0], [1.0, 8.0, 0.0]]), closed=False)
        actual = original.cut_by_plane(
            Plane(point_on_plane=np.array([0.0, 7.5, 0.0]), unit_normal=vg.basis.y)
        )

        np.testing.assert_array_almost_equal(actual.v, expected.v)
        self.assertFalse(actual.closed)

        expected = Polyline(
            np.array(
                [
                    [0.0, 0.0, 0.0],
                    [1.0, 0.0, 0.0],
                    [1.0, 1.0, 0.0],
                    [1.0, 7.0, 0.0],
                    [1.0, 7.5, 0.0],
                ]
            ),
            closed=False,
        )
        actual = original.cut_by_plane(
            Plane(point_on_plane=np.array([0.0, 7.5, 0.0]), unit_normal=vg.basis.neg_y)
        )

        np.testing.assert_array_almost_equal(actual.v, expected.v)
        self.assertFalse(actual.closed)



import numpy as np


class Box(object):
    """
    An axis-aligned cuboid.

    """

    def __init__(self, origin, shape):
        """
        origin: The x, y, z coordinate of the origin.
        shape: A sequence containing the width (dx), height (dy), and
          depth (dz).

        """
        from .._temporary.coercion import as_numeric_array

        self.origin = as_numeric_array(origin, shape=(3,))
        self.shape = as_numeric_array(shape, shape=(3,))
        if any(np.less(self.shape, 0)):
            raise ValueError("Shape should be zero or positive")

    @property
    def ranges(self):
        """
        Return ranges for each coordinate axis as a 3x2 numpy array.

        """
        ranges = np.array([self.origin, self.origin + self.shape]).T
        # ranges is almost, but not quite what we want, since it might
        # include mins which are greater than maxes, and vice versa.
        return np.vstack([ranges.min(axis=1), ranges.max(axis=1)]).T

    @property
    def min_x(self):
        return self.origin[0]

    @property
    def min_y(self):
        return self.origin[1]

    @property
    def min_z(self):
        return self.origin[2]

    @property
    def max_x(self):
        return self.origin[0] + self.shape[0]

    @property
    def max_y(self):
        return self.origin[1] + self.shape[1]

    @property
    def max_z(self):
        return self.origin[2] + self.shape[2]

    @property
    def mid_x(self):
        return self.origin[0] + self.shape[0] / 2

    @property
    def mid_y(self):
        return self.origin[1] + self.shape[1] / 2

    @property
    def mid_z(self):
        return self.origin[2] + self.shape[2] / 2

    @property
    def width(self):
        return self.shape[0]

    @property
    def height(self):
        return self.shape[1]

    @property
    def depth(self):
        return self.shape[2]

    @property
    def center_point(self):
        return self.origin + 0.5 * self.shape

    @property
    def floor_point(self):
        return self.origin + [0.5, 0.0, 0.5] * self.shape

    @property
    def volume(self):
        return np.prod(self.shape)

    @property
    def surface_area(self):
        l, h, w = (
            self.shape
        )  # self.shape is a np.ndarray, which is a sequence. pylint: disable=unpacking-non-sequence
        return 2 * (w * l + h * l + h * w)



import math
from collections import namedtuple
import unittest
import numpy as np
import vg
from .plane import Plane


class DistanceToPlaneTests(unittest.TestCase):
    def test_returns_signed_distances_for_xz_plane_at_origin(self):
        # x-z plane
        normal = np.array([0.0, 1.0, 0.0])
        sample = np.array([0.0, 0.0, 0.0])

        plane = Plane(sample, normal)

        pts = np.array([[500.0, 502.0, 503.0], [-500.0, -501.0, -503.0]])

        expected = np.array([502.0, -501.0])

        np.testing.assert_array_equal(expected, plane.signed_distance(pts))

    def test_returns_unsigned_distances_for_xz_plane_at_origin(self):
        # x-z plane
        normal = np.array([0.0, 1.0, 0.0])
        sample = np.array([0.0, 0.0, 0.0])

        plane = Plane(sample, normal)

        pts = np.array([[500.0, 502.0, 503.0], [-500.0, -501.0, -503.0]])

        expected = np.array([502.0, 501.0])

        np.testing.assert_array_equal(expected, plane.distance(pts))

    def test_returns_signed_distances_for_diagonal_plane(self):
        # diagonal plane @ origin - draw a picture!
        normal = np.array([1.0, 1.0, 0.0])
        normal /= np.linalg.norm(normal)
        sample = np.array([1.0, 1.0, 0.0])

        plane = Plane(sample, normal)

        pts = np.array([[425.0, 425.0, 25.0], [-500.0, -500.0, 25.0]])

        expected = np.array(
            [math.sqrt(2 * (425.0 - 1.0) ** 2), -math.sqrt(2 * (500.0 + 1.0) ** 2)]
        )

        np.testing.assert_array_almost_equal(expected, plane.signed_distance(pts))

    def test_returns_unsigned_distances_for_diagonal_plane_at_origin(self):
        # diagonal plane @ origin - draw a picture!
        normal = np.array([1.0, 1.0, 0.0])
        normal /= np.linalg.norm(normal)

        sample = np.array([0.0, 0.0, 0.0])

        plane = Plane(sample, normal)

        pts = np.array([[425.0, 425.0, 25.0], [-500.0, -500.0, 25.0]])

        expected = np.array([math.sqrt(2 * (425.0 ** 2)), math.sqrt(2 * (500.0 ** 2))])

        np.testing.assert_array_almost_equal(expected, plane.distance(pts))

    def test_returns_sign_for_diagonal_plane(self):
        # diagonal plane @ origin - draw a picture!
        normal = np.array([1.0, 1.0, 0.0])
        normal /= np.linalg.norm(normal)
        sample = np.array([1.0, 1.0, 0.0])

        plane = Plane(sample, normal)

        pts = np.array([[425.0, 425.0, 25.0], [-500.0, -500.0, 25.0]])

        sign = plane.sign(pts)

        expected = np.array([1.0, -1.0])
        np.testing.assert_array_equal(sign, expected)


class TestCanonicalPoint(unittest.TestCase):
    def test_canonical_point(self):
        normal = np.array([1.0, 1.0, 0.0])
        normal /= np.linalg.norm(normal)

        sample = np.array([0.0, 0.0, 0.0])

        plane = Plane(sample, normal)

        np.testing.assert_array_equal(plane.canonical_point, np.array([0.0, 0.0, 0.0]))

        plane = Plane(sample, -normal)

        np.testing.assert_array_equal(plane.canonical_point, np.array([0.0, 0.0, 0.0]))

        normal = np.array([1.0, 7.0, 9.0])
        normal /= np.linalg.norm(normal)

        plane = Plane(sample, normal)

        np.testing.assert_array_equal(plane.canonical_point, np.array([0.0, 0.0, 0.0]))

        plane = Plane(sample, -normal)

        np.testing.assert_array_equal(plane.canonical_point, np.array([0.0, 0.0, 0.0]))

        normal = np.array([1.0, 0.0, 0.0])
        normal /= np.linalg.norm(normal)

        sample = np.array([3.0, 10.0, 20.0])

        plane = Plane(sample, normal)

        np.testing.assert_array_equal(plane.canonical_point, np.array([3, 0.0, 0.0]))

        plane = Plane(sample, -normal)

        np.testing.assert_array_equal(plane.canonical_point, np.array([3, 0.0, 0.0]))

        normal = np.array([1.0, 1.0, 1.0])
        normal /= np.linalg.norm(normal)

        sample = np.array([1.0, 2.0, 10.0])

        plane = Plane(sample, normal)

        np.testing.assert_array_almost_equal(
            plane.canonical_point, np.array([4.333333, 4.333333, 4.333333])
        )

        plane = Plane(sample, -normal)

        np.testing.assert_array_almost_equal(
            plane.canonical_point, np.array([4.333333, 4.333333, 4.333333])
        )


class TestProjectPoint(unittest.TestCase):
    def test_project_point(self):
        plane = Plane(point_on_plane=np.array([0, 10, 0]), unit_normal=vg.basis.y)

        point = np.array([10, 20, -5])

        expected = np.array([10, 10, -5])

        np.testing.assert_array_equal(plane.project_point(point), expected)


class TestPlaneFromPoints(unittest.TestCase):
    def test_plane_from_points(self):
        points = np.array([[1, 1, 1], [-1, 1, 0], [2, 0, 3]])
        plane = Plane.from_points(*points)

        a, b, c, d = plane.equation

        plane_equation_test = [a * x + b * y + c * z + d for x, y, z in points]
        np.testing.assert_array_equal(plane_equation_test, [0, 0, 0])

        projected_points = [plane.project_point(p) for p in points]
        np.testing.assert_array_almost_equal(projected_points, points)

    def test_plane_from_points_order(self):
        points = np.array([[1, 0, 0], [0, math.sqrt(1.25), 0], [-1, 0, 0]])
        plane = Plane.from_points(*points)

        expected_v = np.array([0, 0, 1])
        np.testing.assert_array_equal(plane.normal, expected_v)

    def test_plane_from_points_and_vector(self):
        p1 = np.array([1, 5, 7])
        p2 = np.array([-2, -2, -2])
        v = np.array([1, 0, -1])
        plane = Plane.from_points_and_vector(p1, p2, v)

        points = [p1, p2]
        projected_points = [plane.project_point(p) for p in points]
        np.testing.assert_array_almost_equal(projected_points, points)

        self.assertEqual(np.dot(v, plane.normal), 0)

    def test_fit_from_points(self):
        # Set up a collection of points in the X-Y plane.
        np.random.seed(0)
        points = np.hstack([np.random.random((100, 2)), np.zeros(100).reshape(-1, 1)])
        plane = Plane.fit_from_points(points)

        # The normal vector should be closely aligned with the Z-axis.
        z_axis = np.array([0.0, 0.0, 1.0])
        angle = np.arccos(np.dot(plane.normal, z_axis) / np.linalg.norm(plane.normal))
        self.assertTrue(angle % np.pi < 1e-6)


MockMesh = namedtuple("MockMesh", ["v", "f"])


class PlaneXSectionTests(unittest.TestCase):
    def setUp(self):
        self.box_mesh = MockMesh(
            v=np.array(
                [
                    [0.5, -0.5, 0.5, -0.5, 0.5, -0.5, 0.5, -0.5],
                    [0.5, 0.5, -0.5, -0.5, 0.5, 0.5, -0.5, -0.5],
                    [0.5, 0.5, 0.5, 0.5, -0.5, -0.5, -0.5, -0.5],
                ]
            ).T,
            f=np.array(
                [
                    [0, 1, 2],
                    [3, 2, 1],
                    [0, 2, 4],
                    [6, 4, 2],
                    [0, 4, 1],
                    [5, 1, 4],
                    [7, 5, 6],
                    [4, 6, 5],
                    [7, 6, 3],
                    [2, 3, 6],
                    [7, 3, 5],
                    [1, 5, 3],
                ]
            ),
        )

    def double_mesh(self, mesh, shift=[2.0, 0.0, 0.0]):
        other_mesh = MockMesh(v=mesh.v + np.array(shift), f=mesh.f)
        two_meshes = MockMesh(
            v=np.vstack((mesh.v, other_mesh.v)),
            f=np.vstack((mesh.f, other_mesh.f + mesh.v.shape[0])),
        )
        return two_meshes

    def test_line_plane_intersection(self):
        # x-z plane
        normal = np.array([0.0, 1.0, 0.0])
        sample = np.array([0.0, 0.0, 0.0])

        plane = Plane(sample, normal)
        self.assertIsNone(
            plane.line_xsection(pt=[0.0, -1.0, 0.0], ray=[1.0, 0.0, 0.0])
        )  # non-intersecting
        self.assertIsNone(
            plane.line_xsection(pt=[0.0, 0.0, 0.0], ray=[1.0, 0.0, 0.0])
        )  # coplanar
        np.testing.assert_array_equal(
            plane.line_xsection(pt=[0.0, -1.0, 0.0], ray=[0.0, 1.0, 0.0]),
            [0.0, 0.0, 0.0],
        )
        np.testing.assert_array_equal(
            plane.line_xsection(pt=[0.0, -1.0, 0.0], ray=[1.0, 1.0, 0.0]),
            [1.0, 0.0, 0.0],
        )

    def test_line_plane_intersections(self):
        # x-z plane
        normal = np.array([0.0, 1.0, 0.0])
        sample = np.array([0.0, 0.0, 0.0])

        plane = Plane(sample, normal)
        pts = np.array(
            [[0.0, -1.0, 0.0], [0.0, 0.0, 0.0], [0.0, -1.0, 0.0], [0.0, -1.0, 0.0]]
        )
        rays = np.array(
            [[1.0, 0.0, 0.0], [1.0, 0.0, 0.0], [0.0, 1.0, 0.0], [1.0, 1.0, 0.0]]
        )
        expected = np.array(
            [
                [np.nan, np.nan, np.nan],
                [np.nan, np.nan, np.nan],
                [0.0, 0.0, 0.0],
                [1.0, 0.0, 0.0],
            ]
        )
        intersections, is_intersecting = plane.line_xsections(pts, rays)
        np.testing.assert_array_equal(intersections, expected)
        np.testing.assert_array_equal(is_intersecting, [False, False, True, True])

    def test_line_segment_plane_intersection(self):
        # x-z plane
        normal = np.array([0.0, 1.0, 0.0])
        sample = np.array([0.0, 0.0, 0.0])

        plane = Plane(sample, normal)
        self.assertIsNone(
            plane.line_segment_xsection([0.0, -1.0, 0.0], [1.0, -1.0, 0.0])
        )  # non-intersecting
        self.assertIsNone(
            plane.line_segment_xsection([0.0, 0.0, 0.0], [1.0, 0.0, 0.0])
        )  # coplanar
        np.testing.assert_array_equal(
            plane.line_segment_xsection([0.0, -1.0, 0.0], [0.0, 1.0, 0.0]),
            [0.0, 0.0, 0.0],
        )
        np.testing.assert_array_equal(
            plane.line_segment_xsection([0.0, -1.0, 0.0], [2.0, 1.0, 0.0]),
            [1.0, 0.0, 0.0],
        )
        self.assertIsNone(
            plane.line_segment_xsection([0.0, 1.0, 0.0], [0.0, 2.0, 0.0])
        )  # line intersecting, but not in segment

    def test_line_segment_plane_intersections(self):
        # x-z plane
        normal = np.array([0.0, 1.0, 0.0])
        sample = np.array([0.0, 0.0, 0.0])

        plane = Plane(sample, normal)
        a = np.array(
            [
                [0.0, -1.0, 0.0],
                [0.0, 0.0, 0.0],
                [0.0, -1.0, 0.0],
                [0.0, -1.0, 0.0],
                [0.0, 1.0, 0.0],
            ]
        )
        b = np.array(
            [
                [1.0, -1.0, 0.0],
                [1.0, 0.0, 0.0],
                [0.0, 1.0, 0.0],
                [2.0, 1.0, 0.0],
                [0.0, 2.0, 0.0],
            ]
        )
        expected = np.array(
            [
                [np.nan, np.nan, np.nan],
                [np.nan, np.nan, np.nan],
                [0.0, 0.0, 0.0],
                [1.0, 0.0, 0.0],
                [np.nan, np.nan, np.nan],
            ]
        )
        intersections, is_intersecting = plane.line_segment_xsections(a, b)
        np.testing.assert_array_equal(intersections, expected)
        np.testing.assert_array_equal(
            is_intersecting, [False, False, True, True, False]
        )



import unittest
import numpy as np
import vg
from .surface_normals import surface_normal


class TestSurfaceNormals(unittest.TestCase):
    def test_surface_normal_single(self):
        points = np.array([[3.0, 0.0, 0.0], [0.0, 3.0, 0.0], [0.0, 0.0, 3.0]])

        np.testing.assert_allclose(
            surface_normal(points), np.array([3 ** -0.5, 3 ** -0.5, 3 ** -0.5])
        )

        np.testing.assert_allclose(
            surface_normal(points, normalize=False), np.array([9.0, 9.0, 9.0])
        )

    def test_surface_normal_vectorized(self):
        from .shapes import create_triangular_prism

        p1 = np.array([3.0, 0.0, 0.0])
        p2 = np.array([0.0, 3.0, 0.0])
        p3 = np.array([0.0, 0.0, 3.0])
        vertices = create_triangular_prism(p1, p2, p3, 1.0)

        expected_normals = vg.normalize(
            np.array(
                [
                    [1.0, 1.0, 1.0],
                    [1.0, 1.0, -2.0],
                    [1.0, 1.0, -2.0],
                    [-2.0, 1.0, 1.0],
                    [-2.0, 1.0, 1.0],
                    [1.0, -2.0, 1.0],
                    [1.0, -2.0, 1.0],
                    [-1.0, -1.0, -1.0],
                ]
            )
        )

        np.testing.assert_allclose(surface_normal(vertices), expected_normals)



def as_numeric_array(a, shape=None, allow_none=False, empty_as_none=False, dtype=None):
    """
    Coerce a array-like input to a numpy ndarray, and validate it.

    a: An array-like input.

    shape: Provide a shape tuple here to validate the shape of the
      array. To require 3 by 1, pass (3,). To require n by 3, pass
      (-1, 3).

    allow_none: When True, and a is None, return None. When False
      and a is None, raise ValueError.

    empty_as_none: For legacy use in Mesh. When a is an empty array,
      print a warning and return None.

    dtype: Coerces to a specific numpy data type, such as np.float64
      or np.uint64. See np.dtype.

    """
    import numpy as np
    from .predicates import isnumericarray

    if a is None:
        if allow_none:
            return None
        else:
            raise ValueError("Value is None")
    a = np.asarray(a, dtype=dtype)
    if not isnumericarray(a):
        raise ValueError("Must be numeric")
    if allow_none and empty_as_none:
        if a.shape == (0,):
            import warnings

            warnings.warn(
                "To clear this value, set it to None instead of []",
                DeprecationWarning,
                stacklevel=2,
            )
            return as_numeric_array(None, shape, allow_none, dtype)
    if shape is not None:
        if len(shape) != len(a.shape):
            raise ValueError("Shape mismatch: expected %s, got %s" % (shape, a.shape))
        if all([exp == -1 or exp == dim for exp, dim in zip(shape, a.shape)]):
            pass
        elif all([exp == -1 or exp == dim for exp, dim in zip(shape, a.T.shape)]):
            a = a.T
        else:
            raise ValueError(
                "Dimension mismatch: expected %s, got %s" % (shape, a.shape)
            )
    return a



import unittest
import numpy as np


class PartitionSegmentOldTests(unittest.TestCase):
    def setUp(self):
        from .segment import partition_segment_old

        self.partition_segment_old = partition_segment_old

    def test_raises_exception_for_invalid_partition_size_type(self):
        p1 = np.array([0.0, 0.0, 0.0])
        p2 = np.array([1.0, 0.0, 0.0])

        self.assertRaises(TypeError, self.partition_segment_old, p1, p2, "foobar")

    def test_raises_exception_for_invalid_partition_size_value(self):
        p1 = np.array([0.0, 0.0, 0.0])
        p2 = np.array([1.0, 0.0, 0.0])

        self.assertRaises(ValueError, self.partition_segment_old, p1, p2, 1)

    def test_returns_partition_for_odd_partition_size(self):
        p1 = np.array([0.0, 0.0, 0.0])
        p2 = np.array([2.0, 0.0, 0.0])

        partition_size = 4

        expected_partition_points = np.array(
            [[0.5, 0.0, 0.0], [1.0, 0.0, 0.0], [1.5, 0.0, 0.0]]
        )

        np.testing.assert_array_almost_equal(
            self.partition_segment_old(p1, p2, partition_size),
            expected_partition_points,
            decimal=7,
        )

    def test_returns_partition_points_for_even_partition_size(self):
        p1 = np.array([0.0, 0.0, 0.0])
        p2 = np.array([1.0, 0.0, 0.0])

        partition_size = 5

        expected_partition_points = np.array(
            [[0.2, 0.0, 0.0], [0.4, 0.0, 0.0], [0.6, 0.0, 0.0], [0.8, 0.0, 0.0]]
        )

        np.testing.assert_array_almost_equal(
            self.partition_segment_old(p1, p2, partition_size),
            expected_partition_points,
            decimal=7,
        )

    def test_returns_partition_points_in_oriented_order(self):
        p1 = np.array([0.0, 0.0, 0.0])
        p2 = np.array([1.0, 0.0, 0.0])

        partition_size = 5

        expected_partition_points = np.array(
            [[0.8, 0.0, 0.0], [0.6, 0.0, 0.0], [0.4, 0.0, 0.0], [0.2, 0.0, 0.0]]
        )

        np.testing.assert_array_almost_equal(
            self.partition_segment_old(p2, p1, partition_size),
            expected_partition_points,
            decimal=7,
        )

    def test_returns_partition_points_for_diagonal_segment(self):
        p1 = np.array([0.0, 0.0, 0.0])
        p2 = np.array([1.0, 1.0, 0.0])

        partition_size = 3

        dist = np.linalg.norm(p2 - p1)
        domain = [(1 / 3.0) * dist, (2 / 3.0) * dist]

        unit_direction = (p2 - p1) / dist

        expected_partition_points = np.array(
            [p1 + scalar * unit_direction for scalar in domain]
        )

        np.testing.assert_array_almost_equal(
            self.partition_segment_old(p1, p2, partition_size),
            expected_partition_points,
            decimal=7,
        )


class PartitionSegmentTests(unittest.TestCase):
    def setUp(self):
        from .segment import partition_segment

        self.partition_segment = partition_segment

    def test_raises_exception_for_invalid_partition_size_type(self):
        p1 = np.array([0.0, 0.0, 0.0])
        p2 = np.array([1.0, 0.0, 0.0])

        self.assertRaises(TypeError, self.partition_segment, p1, p2, "foobar")

    def test_raises_exception_for_invalid_partition_size_value(self):
        p1 = np.array([0.0, 0.0, 0.0])
        p2 = np.array([1.0, 0.0, 0.0])

        self.assertRaises(ValueError, self.partition_segment, p1, p2, 1)

    def test_returns_partition_for_odd_partition_size(self):
        p1 = np.array([0.0, 0.0, 0.0])
        p2 = np.array([2.0, 0.0, 0.0])

        partition_size = 5

        expected_partition_points = np.array(
            [
                [0.0, 0.0, 0.0],
                [0.5, 0.0, 0.0],
                [1.0, 0.0, 0.0],
                [1.5, 0.0, 0.0],
                [2.0, 0.0, 0.0],
            ]
        )

        np.testing.assert_array_almost_equal(
            self.partition_segment(p1, p2, partition_size),
            expected_partition_points,
            decimal=7,
        )

    def test_returns_partition_points_for_even_partition_size(self):
        p1 = np.array([0.0, 0.0, 0.0])
        p2 = np.array([1.0, 0.0, 0.0])

        partition_size = 6

        expected_partition_points = np.array(
            [
                [0.0, 0.0, 0.0],
                [0.2, 0.0, 0.0],
                [0.4, 0.0, 0.0],
                [0.6, 0.0, 0.0],
                [0.8, 0.0, 0.0],
                [1.0, 0.0, 0.0],
            ]
        )

        np.testing.assert_array_almost_equal(
            self.partition_segment(p1, p2, partition_size),
            expected_partition_points,
            decimal=7,
        )

    def test_returns_partition_omitting_endpoint(self):
        p1 = np.array([0.0, 0.0, 0.0])
        p2 = np.array([1.0, 0.0, 0.0])

        partition_size = 5

        expected_partition_points = np.array(
            [
                [0.0, 0.0, 0.0],
                [0.2, 0.0, 0.0],
                [0.4, 0.0, 0.0],
                [0.6, 0.0, 0.0],
                [0.8, 0.0, 0.0],
            ]
        )

        np.testing.assert_array_almost_equal(
            self.partition_segment(p1, p2, partition_size, endpoint=False),
            expected_partition_points,
            decimal=7,
        )


class AddPointsTests(unittest.TestCase):
    def setUp(self):
        from .segment import partition

        self.partition = partition

    def test_adds_points_for_equal_length_line_segments(self):
        v = np.array(
            [
                [0.0, 0.0, 0.0],
                [1.0, 0.0, 0.0],
                [1.0, 1.0, 0.0],
                [1.0, 2.0, 0.0],
                [1.0, 3.0, 0.0],
            ]
        )

        expected = np.array(
            [
                [0.0, 0.0, 0.0],
                [0.2, 0.0, 0.0],
                [0.4, 0.0, 0.0],
                [0.6, 0.0, 0.0],
                [0.8, 0.0, 0.0],
                [1.0, 0.0, 0.0],
                [1.0, 0.2, 0.0],
                [1.0, 0.4, 0.0],
                [1.0, 0.6, 0.0],
                [1.0, 0.8, 0.0],
                [1.0, 1.0, 0.0],
                [1.0, 1.2, 0.0],
                [1.0, 1.4, 0.0],
                [1.0, 1.6, 0.0],
                [1.0, 1.8, 0.0],
                [1.0, 2.0, 0.0],
                [1.0, 2.2, 0.0],
                [1.0, 2.4, 0.0],
                [1.0, 2.6, 0.0],
                [1.0, 2.8, 0.0],
                [1.0, 3.0, 0.0],
            ]
        )

        np.testing.assert_array_almost_equal(self.partition(v), expected)

    def test_adds_points_for_nonequal_arbitrarily_oriented_line(self):
        v = np.array(
            [[0.0, 0.0, 0.0], [1.0, 0.0, 1.0], [2.0, 0.0, 1.0], [2.0, 2.0, 1.0]]
        )

        expected = np.array(
            [
                [0.0, 0.0, 0.0],
                [0.2, 0.0, 0.2],
                [0.4, 0.0, 0.4],
                [0.6, 0.0, 0.6],
                [0.8, 0.0, 0.8],
                [1.0, 0.0, 1.0],
                [1.2, 0.0, 1.0],
                [1.4, 0.0, 1.0],
                [1.6, 0.0, 1.0],
                [1.8, 0.0, 1.0],
                [2.0, 0.0, 1.0],
                [2.0, 0.4, 1.0],
                [2.0, 0.8, 1.0],
                [2.0, 1.2, 1.0],
                [2.0, 1.6, 1.0],
                [2.0, 2.0, 1.0],
            ]
        )

        np.testing.assert_array_almost_equal(self.partition(v), expected)
